# This file is part of cloud-init. See LICENSE file for license information.

import os
import stat
from functools import partial
from textwrap import dedent
from typing import NamedTuple
from unittest import mock
from unittest.mock import patch

import pytest

from cloudinit import ssh_util, util

M_PATH = "cloudinit.ssh_util."


class FakePwEnt(NamedTuple):
    pw_name: str = "UNSET_pw_name"
    pw_passwd: str = "UNSET_w_passwd"
    pw_uid: str = "UNSET_pw_uid"
    pw_gid: str = "UNSET_pw_gid"
    pw_gecos: str = "UNSET_pw_gecos"
    pw_dir: str = "UNSET_pw_dir"
    pw_shell: str = "UNSET_pw_shell"


def mock_get_owner(updated_permissions, value):
    try:
        return updated_permissions[value][0]
    except ValueError:
        return util.get_owner(value)


def mock_get_group(updated_permissions, value):
    try:
        return updated_permissions[value][1]
    except ValueError:
        return util.get_group(value)


def mock_get_user_groups(username):
    return username


def mock_get_permissions(updated_permissions, value):
    try:
        return updated_permissions[value][2]
    except ValueError:
        return util.get_permissions(value)


def mock_getpwnam(users, username):
    return users[username]


# Do not use these public keys, most of them are fetched from
# the testdata for OpenSSH, and their private keys are available
# https://github.com/openssh/openssh-portable/tree/master/regress/unittests/sshkey/testdata
VALID_CONTENT = {
    "ecdsa": (
        "AAAAE2VjZHNhLXNoYTItbmlzdHAyNTYAAAAIbmlzdHAyNTYAAABBBITrGBB3cgJ"
        "J7fPxvtMW9H3oRisNpJ3OAslxZeyP7I0A9BPAW0RQIwHVtVnM7zrp4nI+JLZov/"
        "Ql7lc2leWL7CY="
    ),
    "rsa": (
        "AAAAB3NzaC1yc2EAAAABIwAAAQEA3I7VUf2l5gSn5uavROsc5HRDpZdQueUq5oz"
        "emNSj8T7enqKHOEaFoU2VoPgGEWC9RyzSQVeyD6s7APMcE82EtmW4skVEgEGSbD"
        "c1pvxzxtchBj78hJP6Cf5TCMFSXw+Fz5rF1dR23QDbN1mkHs7adr8GW4kSWqU7Q"
        "7NDwfIrJJtO7Hi42GyXtvEONHbiRPOe8stqUly7MvUoN+5kfjBM8Qqpfl2+FNhT"
        "YWpMfYdPUnE7u536WqzFmsaqJctz3gBxH9Ex7dFtrxR4qiqEr9Qtlu3xGn7Bw07"
        "/+i1D+ey3ONkZLN+LQ714cgj8fRS4Hj29SCmXp5Kt5/82cD/VN3NtHw=="
    ),
    "ed25519": (
        "AAAAC3NzaC1lZDI1NTE5AAAAIA1J77+CrJ8p6/vWCEzuylqJNMHUP/XmeYyGVWb8lnDd"
    ),
    "ecdsa-sha2-nistp256-cert-v01@openssh.com": (
        "AAAAKGVjZHNhLXNoYTItbmlzdHAyNTYtY2VydC12MDFAb3BlbnNzaC5jb20AAAA"
        "gQIfwT/+UX68/hlKsdKuaOuAVB6ftTg03SlP/uH4OBEwAAAAIbmlzdHAyNTYAAA"
        "BBBEjA0gjJmPM6La3sXyfNlnjilvvGY6I2M8SvJj4o3X/46wcUbPWTaj4RF3EXw"
        "HvNxplYBwdPlk2zEecvf9Cs2BMAAAAAAAAAAAAAAAEAAAAYa2V5cy9lY2RzYS1z"
        "aGEyLW5pc3RwMjU2AAAAAAAAAAAAAAAA//////////8AAAAAAAAAggAAABVwZXJ"
        "taXQtWDExLWZvcndhcmRpbmcAAAAAAAAAF3Blcm1pdC1hZ2VudC1mb3J3YXJkaW"
        "5nAAAAAAAAABZwZXJtaXQtcG9ydC1mb3J3YXJkaW5nAAAAAAAAAApwZXJtaXQtc"
        "HR5AAAAAAAAAA5wZXJtaXQtdXNlci1yYwAAAAAAAAAAAAAAaAAAABNlY2RzYS1z"
        "aGEyLW5pc3RwMjU2AAAACG5pc3RwMjU2AAAAQQRH6Y9Q1+ocQ8ETKW3LjQqtxg7"
        "OuSSDacxmmQatQVaIawwjCbmntyEAqmVj3v9ElDSXnO5m7TyYMBQu4+vsh76RAA"
        "AAZQAAABNlY2RzYS1zaGEyLW5pc3RwMjU2AAAASgAAACEA47Cl2MMhr+glPGuxx"
        "2tM3QXkDcwdP0SxSEW5yy4XV5oAAAAhANNMm1cdVlAt3hmycQgdD82zPlg5YvVO"
        "iN0SQTbgVD8i"
    ),
    "ecdsa-sha2-nistp256": (
        "AAAAE2VjZHNhLXNoYTItbmlzdHAyNTYAAAAIbmlzdHAyNTYAAABBBEjA0gjJmPM"
        "6La3sXyfNlnjilvvGY6I2M8SvJj4o3X/46wcUbPWTaj4RF3EXwHvNxplYBwdPlk"
        "2zEecvf9Cs2BM="
    ),
    "ecdsa-sha2-nistp384-cert-v01@openssh.com": (
        "AAAAKGVjZHNhLXNoYTItbmlzdHAzODQtY2VydC12MDFAb3BlbnNzaC5jb20AAAA"
        "grnSvDsK1EnCZndO1IyGWcGkVgVSkPWi/XO2ybPFyLVUAAAAIbmlzdHAzODQAAA"
        "BhBAaYSQs+8TT0Tzciy0dorwhur6yzOGUrYQ6ueUQYWbE7eNdHmhsVrlpGPgSaY"
        "ByhXtAJiPOMqLU5h0eb3sCtM3ek4NvjXFTGTqPrrxJI6q0OsgrtkGE7UM9ZsfMm"
        "7q6BOAAAAAAAAAAAAAAAAQAAABhrZXlzL2VjZHNhLXNoYTItbmlzdHAzODQAAAA"
        "AAAAAAAAAAAD//////////wAAAAAAAACCAAAAFXBlcm1pdC1YMTEtZm9yd2FyZG"
        "luZwAAAAAAAAAXcGVybWl0LWFnZW50LWZvcndhcmRpbmcAAAAAAAAAFnBlcm1pd"
        "C1wb3J0LWZvcndhcmRpbmcAAAAAAAAACnBlcm1pdC1wdHkAAAAAAAAADnBlcm1p"
        "dC11c2VyLXJjAAAAAAAAAAAAAACIAAAAE2VjZHNhLXNoYTItbmlzdHAzODQAAAA"
        "IbmlzdHAzODQAAABhBLWbubcMzcWc7lMTCMGVXZlaVvUOHLjpr6SOOScFFrd8K9"
        "Gl8nYELST5HZ1gym65m+MG6/tbrUWIY/flLWNIe+WtqxrdPPGdIhFruCwNw2peZ"
        "SbQOa/o3AGnJ/vO6EKEGAAAAIQAAAATZWNkc2Etc2hhMi1uaXN0cDM4NAAAAGkA"
        "AAAxAL10JHd5bvnbpD+fet/k1YE1BEIrqGXaoIIJ9ReE5H4nTK1uQJzMD7+wwGK"
        "RVYqYQgAAADAiit0UCMDAUbjD+R2x4LvU3x/t8G3sdqDLRNfMRpjZpvcS8AwC+Y"
        "VFVSQNn0AyzW0="
    ),
    "ecdsa-sha2-nistp384": (
        "AAAAE2VjZHNhLXNoYTItbmlzdHAzODQAAAAIbmlzdHAzODQAAABhBAaYSQs+8TT"
        "0Tzciy0dorwhur6yzOGUrYQ6ueUQYWbE7eNdHmhsVrlpGPgSaYByhXtAJiPOMqL"
        "U5h0eb3sCtM3ek4NvjXFTGTqPrrxJI6q0OsgrtkGE7UM9ZsfMm7q6BOA=="
    ),
    "ecdsa-sha2-nistp521-cert-v01@openssh.com": (
        "AAAAKGVjZHNhLXNoYTItbmlzdHA1MjEtY2VydC12MDFAb3BlbnNzaC5jb20AAAA"
        "gGmRzkkMvRFk1V5U3m3mQ2nfW20SJVXk1NKnT5iZGDcEAAAAIbmlzdHA1MjEAAA"
        "CFBAHosAOHAI1ZkerbKYQ72S6uit1u77PCj/OalZtXgsxv0TTAZB273puG2X94C"
        "Q8yyNHcby87zFZHdv5BSKyZ/cyREAAeiAcSakop9VS3+bUfZpEIqwBZXarwUjnR"
        "nxprkcQ0rfCCdagkGZr/OA7DemK2D8tKLTHsKoEEWNImo6/pXDkFxAAAAAAAAAA"
        "AAAAAAQAAABhrZXlzL2VjZHNhLXNoYTItbmlzdHA1MjEAAAAAAAAAAAAAAAD///"
        "///////wAAAAAAAACCAAAAFXBlcm1pdC1YMTEtZm9yd2FyZGluZwAAAAAAAAAXc"
        "GVybWl0LWFnZW50LWZvcndhcmRpbmcAAAAAAAAAFnBlcm1pdC1wb3J0LWZvcndh"
        "cmRpbmcAAAAAAAAACnBlcm1pdC1wdHkAAAAAAAAADnBlcm1pdC11c2VyLXJjAAA"
        "AAAAAAAAAAACsAAAAE2VjZHNhLXNoYTItbmlzdHA1MjEAAAAIbmlzdHA1MjEAAA"
        "CFBAC6hFVXM1XEg/7qKkp5sLZuANGQVW88b5pPn2ZcK0td9IQstLH6BwWuZ6MPE"
        "ogiDlvx9HD1BaKGBBfkxgOY8NGFzQHbjU9eTWH3gt0RATDbZsij1pSkFPnAXdU9"
        "SjfogYloI2xdHaTCgWp3zgsUV+BBQ0QGGv2MqqcOmrF0f5YEJeOffAAAAKcAAAA"
        "TZWNkc2Etc2hhMi1uaXN0cDUyMQAAAIwAAABCAT+vSOYPuYVTDopDW08576d5Sb"
        "edXQMOu1op4CQIm98VKtAXvu5dfioi5VYAqpte8M+UxEMOMiQWJp+U9exYf6LuA"
        "AAAQgEzkIpX3yKXPaPcK17mNx40ujEDitm4ARmbhAge0sFhZtf7YIgI55b6vkI8"
        "JvMJkzQCBF1cpNOaIpVh1nFZNBphMQ=="
    ),
    "ecdsa-sha2-nistp521": (
        "AAAAE2VjZHNhLXNoYTItbmlzdHA1MjEAAAAIbmlzdHA1MjEAAACFBAHosAOHAI1"
        "ZkerbKYQ72S6uit1u77PCj/OalZtXgsxv0TTAZB273puG2X94CQ8yyNHcby87zF"
        "ZHdv5BSKyZ/cyREAAeiAcSakop9VS3+bUfZpEIqwBZXarwUjnRnxprkcQ0rfCCd"
        "agkGZr/OA7DemK2D8tKLTHsKoEEWNImo6/pXDkFxA=="
    ),
    "sk-ecdsa-sha2-nistp256-cert-v01@openssh.com": (
        "AAAAIHNzaC1lZDI1NTE5LWNlcnQtdjAxQG9wZW5zc2guY29tAAAAIIxzuxl4z3u"
        "wAIslne8Huft+1n1IhHAlNbWZkQyyECCGAAAAIFOG6kY7Rf4UtCFvPwKgo/BztX"
        "ck2xC4a2WyA34XtIwZAAAAAAAAAAgAAAACAAAABmp1bGl1cwAAABIAAAAFaG9zd"
        "DEAAAAFaG9zdDIAAAAANowB8AAAAABNHmBwAAAAAAAAAAAAAAAAAAAAMwAAAAtz"
        "c2gtZWQyNTUxOQAAACBThupGO0X+FLQhbz8CoKPwc7V3JNsQuGtlsgN+F7SMGQA"
        "AAFMAAAALc3NoLWVkMjU1MTkAAABABGTn+Bmz86Ajk+iqKCSdP5NClsYzn4alJd"
        "0V5bizhP0Kumc/HbqQfSt684J1WdSzih+EjvnTgBhK9jTBKb90AQ=="
    ),
    "sk-ecdsa-sha2-nistp256@openssh.com": (
        "AAAAInNrLWVjZHNhLXNoYTItbmlzdHAyNTZAb3BlbnNzaC5jb20AAAAIbmlzdHA"
        "yNTYAAABBBIELQJ2DgvaX1yQlKFokfWM2suuaCFI2qp0eJodHyg6O4ifxc3XpRK"
        "d1OS8dNYQtE/YjdXSrA+AOnMF5ns2Nkx4AAAAEc3NoOg=="
    ),
    "sk-ssh-ed25519-cert-v01@openssh.com": (
        "AAAAIHNzaC1lZDI1NTE5LWNlcnQtdjAxQG9wZW5zc2guY29tAAAAIIxzuxl4z3u"
        "wAIslne8Huft+1n1IhHAlNbWZkQyyECCGAAAAIFOG6kY7Rf4UtCFvPwKgo/BztX"
        "ck2xC4a2WyA34XtIwZAAAAAAAAAAgAAAACAAAABmp1bGl1cwAAABIAAAAFaG9zd"
        "DEAAAAFaG9zdDIAAAAANowB8AAAAABNHmBwAAAAAAAAAAAAAAAAAAAAMwAAAAtz"
        "c2gtZWQyNTUxOQAAACBThupGO0X+FLQhbz8CoKPwc7V3JNsQuGtlsgN+F7SMGQA"
        "AAFMAAAALc3NoLWVkMjU1MTkAAABABGTn+Bmz86Ajk+iqKCSdP5NClsYzn4alJd"
        "0V5bizhP0Kumc/HbqQfSt684J1WdSzih+EjvnTgBhK9jTBKb90AQ=="
    ),
    "sk-ssh-ed25519@openssh.com": (
        "AAAAGnNrLXNzaC1lZDI1NTE5QG9wZW5zc2guY29tAAAAICFo/k5LU8863u66YC9"
        "eUO2170QduohPURkQnbLa/dczAAAABHNzaDo="
    ),
    "ssh-ed25519-cert-v01@openssh.com": (
        "AAAAIHNzaC1lZDI1NTE5LWNlcnQtdjAxQG9wZW5zc2guY29tAAAAIIxzuxl4z3u"
        "wAIslne8Huft+1n1IhHAlNbWZkQyyECCGAAAAIFOG6kY7Rf4UtCFvPwKgo/BztX"
        "ck2xC4a2WyA34XtIwZAAAAAAAAAAgAAAACAAAABmp1bGl1cwAAABIAAAAFaG9zd"
        "DEAAAAFaG9zdDIAAAAANowB8AAAAABNHmBwAAAAAAAAAAAAAAAAAAAAMwAAAAtz"
        "c2gtZWQyNTUxOQAAACBThupGO0X+FLQhbz8CoKPwc7V3JNsQuGtlsgN+F7SMGQA"
        "AAFMAAAALc3NoLWVkMjU1MTkAAABABGTn+Bmz86Ajk+iqKCSdP5NClsYzn4alJd"
        "0V5bizhP0Kumc/HbqQfSt684J1WdSzih+EjvnTgBhK9jTBKb90AQ=="
    ),
    "ssh-ed25519": (
        "AAAAC3NzaC1lZDI1NTE5AAAAIFOG6kY7Rf4UtCFvPwKgo/BztXck2xC4a2WyA34XtIwZ"
    ),
    "ssh-rsa-cert-v01@openssh.com": (
        "AAAAHHNzaC1yc2EtY2VydC12MDFAb3BlbnNzaC5jb20AAAAg98LhS2EHxLOWCLo"
        "pZPwHdg/RJXusnkOqQXSc9R7aITkAAAADAQABAAAAgQDLV5lUTt7FrADseB/CGh"
        "EZzpoojjEW5y8+ePvLppmK3MmMI18ud6vxzpK3bwZLYkVSyfJYI0HmIuGhdu7yM"
        "rW6wb84gbq8C31Xoe9EORcIUuGSvDKdNSM1SjlhDquRblDFB8kToqXyx1lqrXec"
        "XylxIUOL0jE+u0rU1967pDJx+wAAAAAAAAAFAAAAAgAAAAZqdWxpdXMAAAASAAA"
        "ABWhvc3QxAAAABWhvc3QyAAAAADaMAfAAAAAATR5gcAAAAAAAAAAAAAAAAAAAAD"
        "MAAAALc3NoLWVkMjU1MTkAAAAgU4bqRjtF/hS0IW8/AqCj8HO1dyTbELhrZbIDf"
        "he0jBkAAABTAAAAC3NzaC1lZDI1NTE5AAAAQI3QGlUCzC07KorupxpDkkGy6tni"
        "aZ8EvBflzvv+itXWNchGvfUeHmVT6aX0sRqehdz/lR+GmXRoZBhofwh0qAM="
    ),
    "ssh-rsa": (
        "AAAAB3NzaC1yc2EAAAADAQABAAAAgQDLV5lUTt7FrADseB/CGhEZzpoojjEW5y8"
        "+ePvLppmK3MmMI18ud6vxzpK3bwZLYkVSyfJYI0HmIuGhdu7yMrW6wb84gbq8C3"
        "1Xoe9EORcIUuGSvDKdNSM1SjlhDquRblDFB8kToqXyx1lqrXecXylxIUOL0jE+u"
        "0rU1967pDJx+w=="
    ),
    "ssh-xmss-cert-v01@openssh.com": (
        "AAAAHXNzaC14bXNzLWNlcnQtdjAxQG9wZW5zc2guY29tAAAAIM2UD0IH+Igsekq"
        "xjTO5f36exX4WGRMCtDGPjwfbXblxAAAAFVhNU1NfU0hBMi0yNTZfVzE2X0gxMA"
        "AAAEDI83/K5JMOy0BMJgQypRdz35ApAnoQinMJ8ZMoZPaEJF8Z4rANQlfzaAXum"
        "N3RDU5CGIUGGw+WJ904G/wwEq9CAAAAAAAAAAAAAAABAAAACWtleXMveG1zcwAA"
        "AAAAAAAAAAAAAP//////////AAAAAAAAAIIAAAAVcGVybWl0LVgxMS1mb3J3YXJ"
        "kaW5nAAAAAAAAABdwZXJtaXQtYWdlbnQtZm9yd2FyZGluZwAAAAAAAAAWcGVybW"
        "l0LXBvcnQtZm9yd2FyZGluZwAAAAAAAAAKcGVybWl0LXB0eQAAAAAAAAAOcGVyb"
        "Wl0LXVzZXItcmMAAAAAAAAAAAAAAHUAAAAUc3NoLXhtc3NAb3BlbnNzaC5jb20A"
        "AAAVWE1TU19TSEEyLTI1Nl9XMTZfSDEwAAAAQA+irIyT2kaOd07YWZT/QItzNBZ"
        "kUYwnqZJihQ7BxuyiDP4HEFbnfYnnIZXx9Asyi7vDyZRvi+AMSOzmMSq4JnkAAA"
        "ngAAAAFHNzaC14bXNzQG9wZW5zc2guY29tAAAJxAAAAAAFjaKTDc+7Hu2uFGIab"
        "3NAku8HbbGtrq/uGXOxmqxu4RaLqmwofl5iXk3nMwWEhQAb99vAc9D9ZFtfxJO4"
        "STYUTjbj4BxToov/uvbYfE5VeO6sMvkGglgh9YHkCTAItsG8EmGT1SIPfKYzLlN"
        "jvUlbcv0PaPFMJ0wzS9mNfuRf+KUhf3dxQ6zaMrBH3KEJ8Me2kNjhnh6rNPROeI"
        "N+IcStSKsydYuiySGKS/orsH38XysuK5QqLizbHJY3cqLbkW9LsIijb+pfEJh4Y"
        "bOoAbraWAv9ySnWCyRhvw2x8uJ0ZM+p5WSRiZfB3JxCpOhHgiKa9TdmdjnAtnED"
        "zqKOj/gM7y9mesn5ydQI0bENOGymlw0ThUGKbXMxn87Hc9dDPURUBmoO3NGjPDf"
        "7meS39A1ZEGtCe/pbZU9iwxqGx4wJYvB4lutRP2tYC1pA6hjQCcHibvxl5iqj+1"
        "jRjwPr8dbTm4PdETW/7JDSVQXKjxOT0kRLHLelJNeviGx5zSHR5PtnUP3nOBMme"
        "hk9DwcQW9vfKeWSnu9CMnF8xvYJxoPKQwmz0TKo+YVOUnc9/Ma+Ykseof9/W+rk"
        "USQGELc4x7XE5XBKYZZP2PmtxirQ3qTWFw+CeTX2Oa+jPYkzOa7jgmHJ3Fi9Xqw"
        "3L844vRl97e28GmwS0M1SXH+ohES0mO4EcrGh5OLyXBaRTV5QMo+4Bg6FH/HwEn"
        "gG1mdEOAqvctK2QC70c4lHGzfexqwQ2U6WUADPcd/BLOE8Noj1EiXYwZrSA1okZ"
        "FYnS/b89Uo51D2FE4A33V4gcxAglGzVNtrPulkguNT9B4jjNgdIwkTBL9k3ujkG"
        "og6pyYjZ0J5Jp5XPBn+y0LqrpOdZijzrc1OJbX59tTeIbDkM7Fw8As4a03hQPDU"
        "FTOdyMHgLnuLhLXOcqIjvW5axZL/Kx3UET8wrSHizPoa6NErCG4v5mC2M4kBSOW"
        "In1QV27QMaHkL/ZAa3mPsW5iFZtOVEGzw2BW4MZs0qOrcloCENZzOHiMBroKEkH"
        "AbzX6D1FLwml2JpXq4JXlCrdIiFm4+co5ygnWPqb4QGzMlcbjW/x/A16TthNuok"
        "wwlmK5ndKZ76LahyGKEwx2Nv0D+0xilEC1EldtiYRdBNlcGbU/A5EhH5bQ9KVIH"
        "wjWm35pRPLl5224//nqvQKhwFCn9otsR35XHXev3IQ0or3HmQxIvSDOwir1l66z"
        "FFrkyHMWexoucbTBxw1MN3hLb247lcVYJ5+hspJgyoYbfR5RkQVDzhpzskogP7l"
        "K5t0bphu+f+hpvrca7DAiiIZkcR4R1UUQoRnJPRXyXOxlxwS10b51cP9p9jzvZj"
        "d2LUs8yx1KXWSxNHo6WmtYONNaUfdX2OB5+QCvPULfLfFeBrqpX6Yp5wQMM5Cup"
        "k8FEfV07eEgQkVE9nDGKHglWo3kUdOF+XCqWAnXn0b/2bNS9/SSAz6gB1GTFcN/"
        "QsFGlC0QgbCJbQ7LQM6hilRWupWvN5zZ/+HJyyRHuSs5VnQnKiGbIa6AIhx7mP7"
        "8T82gKjU3mHLJWMGKcT3cY8R958Gs+w4OT71VJRMw3kK6qk02WCbD5OtbFeC6ib"
        "KRJKdLK3BzjVs/Fzu3mHVucVby3jpvG1Z8HKspKFhvV7gjFEPu8qHKi4MdAlif/"
        "KakyPk8yZB/dMfaxh7Kv/WpJuSwWNs7RNh29e+ZG+POxqRPWiHqiVw7P17a4dN7"
        "nkVOawdBEyxI4NAY+4zW+0r0bAy6zNBitBvkq3IXfr3De6Upex52sPHvK04PXoV"
        "RI6gjnpPSbLLjpSpcHPKgB7DWefLfhd63BUQbc57D8zm8Jd6qtmzcSKn+wz5/zT"
        "0I6v9I4a+DOjjyqpPpzzNU76pt+Y8SuBgHzMm1vcAdNWlbQrqtScvm0T9AkYni6"
        "47vSh77uwRZKDtMCMSU151tVUavXhtLYLZ6/ll5NhMXkkx8//i7pk1OBjN5LHVQ"
        "0QeimRmavlXU1dJ2rwsFAV+9dDdJXUNOq3VLTo9FrbOzZiWtzzjkJpVJAFREnBn"
        "yIDBK5AXtXE1RzfzaBHzbI2e2kO3t+CSNLWYMFYHBDqaeICYQ9+I9aO/8hnzVSo"
        "fp+8IfWO8iJhppqynUniicW2oCzrn4oczzYNEjImt8CGY7g90GxWfX+ZgXMJfy/"
        "bQiFQL3dZvVypDHEbFoIGz+sxkL83xrP4MZV1V9Wwa64lDXYv01Kp4kQXmmnAZY"
        "KlxBoWqYDXLeLLguSOZxDSCIDpd+YPm39wQ3wOysHW2fmsWtp6FPPlQRUYjsGIP"
        "lfrkJzpoeaPKDtF1m+mOULfEh9kvTKCmKRi385T9ON39D97eWqaM4CCfUGImvdR"
        "DlZLXvjmaAh5BVJ8VJxk75OkP14vWFFlTMv0/k4BYLDKsrNqCREC/G9nQBGcD2D"
        "CLwC2zPNaX2Y9dnyDs2csjN1ibsYttUMnXMgBcnCOkIkVS496Bpc0jQMf35GUgb"
        "PSyliwqCoXjEBP/2eyq0VLFKQ0fXGsHWvElT+Y/7RYNTiYVWttFMxN5H/2EGcgn"
        "lfNHLpQvXH9u/3YminS9GX30hQ7jFhpHXxkK8gZ1mpHL9K3pfKS3lG6EF9wQ23O"
        "qS8m995SG3dp3MzmywxXen/ukXx6bDiEl5VaOvdRUcbhr5Eb3exVDfdWiaJdTYF"
        "WfIfJOWx88drB3J9vFwjmuaoNEOjFsoNAMYthYOxXraXaJblvmUKz6tJ3T8/G7x"
        "B9QGYNBsOqBolKoKHBtsWCosLdWhEZr9VFFh2AJrOW1fx24CIkHnvfTtwYORvQq"
        "Ckuq2bZS1EOdsFkU/X5gwPl6gSUTNhV3IooXkBFL3iBEbfZ6JpQHVVyIuNWjIyN"
        "b2liCn9Nn0VHeNMMRLl7uyw4eKlOX2ogom8SLvihYxcJoqlCwtehpLsKsU4iwME"
        "PmDteW5GBGf4GbnqPFkpIT5ed1jGhdZt/dpsp+v6QhYH1uX4pPxdkdnuc84/yb9"
        "k4SQdKBJ+l3KZkfIxApNWOZqicJfz/eWwS/15hiamRKRuiiUV2zS1V+l8bV7g9O"
        "gy5scPBMONxtfFlGEKikZKurFmzboCOGQKRBEUCpsY44IAp443h59pQdVIb0YAS"
        "kfp2xKHwYij6ELRNdH5MrlFa3bNTskGO4k5XDR4cl/Sma2SXgBKb5XjTtlNmCQG"
        "Gv6lOW7pGXNhs5wfd8K9Ukm6KeLTIlYn1iiKM37YQpa+4JQYljCYhumbqNCkPTZ"
        "rNYClh8fQEQ8XuOCDpomMWu58YOTfbZNMDWs/Ou7RfCjX+VNwjPShDK9joMwWKc"
        "Jy3QalZbaoWtcyyvXxR2sqhVR9F7Cmasq4="
    ),
    "ssh-xmss@openssh.com": (
        "AAAAFHNzaC14bXNzQG9wZW5zc2guY29tAAAAFVhNU1NfU0hBMi0yNTZfVzE2X0g"
        "xMAAAAECqptWnK94d+Sj2xcdTu8gz+75lawZoLSZFqC5IhbYuT/Z3oBZCim6yt+"
        "HAmk6MKldl3Fg+74v4sR/SII0I0Jv/"
    ),
}

KEY_TYPES = list(VALID_CONTENT.keys())

TEST_OPTIONS = (
    "no-port-forwarding,no-agent-forwarding,no-X11-forwarding,"
    'command="echo \'Please login as the user "ubuntu" rather than the'
    'user "root".\';echo;sleep 10"'
)


class TestAuthKeyLineParser:
    @pytest.mark.parametrize("with_options", [True, False])
    @pytest.mark.parametrize("with_comment", [True, False])
    @pytest.mark.parametrize("ktype", KEY_TYPES)
    def test_parse(self, ktype, with_comment, with_options):
        content = VALID_CONTENT[ktype]
        comment = "user-%s@host" % ktype
        options = TEST_OPTIONS

        line_args = []
        if with_options:
            line_args.append(options)
        line_args.extend(
            [
                ktype,
                content,
            ]
        )
        if with_comment:
            line_args.append(comment)
        line = " ".join(line_args)

        key = ssh_util.AuthKeyLineParser().parse(line)

        assert key.base64 == content
        assert key.keytype == ktype
        if with_options:
            assert key.options == options
        else:
            assert key.options is None
        if with_comment:
            assert key.comment == comment
        else:
            assert key.comment == ""

    def test_parse_with_options_passed_in(self):
        # test key line with key type and base64 only
        parser = ssh_util.AuthKeyLineParser()

        baseline = " ".join(("rsa", VALID_CONTENT["rsa"], "user@host"))
        myopts = "no-port-forwarding,no-agent-forwarding"

        key = parser.parse("allowedopt" + " " + baseline)
        assert key.options == "allowedopt"

        key = parser.parse("overridden_opt " + baseline, options=myopts)
        assert key.options == myopts

    def test_parse_invalid_keytype(self):
        parser = ssh_util.AuthKeyLineParser()
        key = parser.parse(" ".join(["badkeytype", VALID_CONTENT["rsa"]]))

        assert not key.valid()


class TestUpdateAuthorizedKeys:
    @pytest.mark.parametrize(
        "new_entries",
        [
            (
                [
                    " ".join(("rsa", VALID_CONTENT["rsa"], "new_comment1")),
                ]
            ),
            pytest.param(
                [
                    " ".join(("rsa", VALID_CONTENT["rsa"], "new_comment1")),
                    "xxx-invalid-thing1",
                    "xxx-invalid-blob2",
                ],
                id="skip-invalid-entries",
            ),
        ],
    )
    def test_new_keys_replace(self, new_entries):
        """new entries with the same base64 should replace old."""
        orig_entries = [
            " ".join(("rsa", VALID_CONTENT["rsa"], "orig_comment1")),
            " ".join(("ecdsa", VALID_CONTENT["ecdsa"], "orig_comment2")),
        ]

        expected = "\n".join([new_entries[0], orig_entries[1]]) + "\n"

        parser = ssh_util.AuthKeyLineParser()
        found = ssh_util.update_authorized_keys(
            [parser.parse(p) for p in orig_entries],
            [parser.parse(p) for p in new_entries],
        )

        assert expected == found


@mock.patch(M_PATH + "util.load_text_file")
@mock.patch(M_PATH + "os.path.isfile")
class TestParseSSHConfig:
    @pytest.mark.parametrize(
        "is_file, file_content",
        [
            pytest.param(True, ("",), id="empty-file"),
            pytest.param(False, IOError, id="not-a-file"),
        ],
    )
    def test_dummy_file(self, m_is_file, m_load_file, is_file, file_content):
        m_is_file.return_value = is_file
        m_load_file.side_effect = file_content
        ret = ssh_util.parse_ssh_config("notmatter")
        assert [] == ret

    @pytest.mark.parametrize(
        "file_content",
        [
            pytest.param(["# This is a comment"], id="comment_line"),
            pytest.param(
                ["# This is a comment", "# This is another comment"],
                id="two-comment_lines",
            ),
        ],
    )
    def test_comment_line(self, m_is_file, m_load_file, file_content):
        m_is_file.return_value = True
        m_load_file.return_value = "\n".join(file_content)
        ret = ssh_util.parse_ssh_config("some real file")
        assert len(file_content) == len(ret)
        assert file_content[0] == ret[0].line

    def test_blank_lines(self, m_is_file, m_load_file):
        m_is_file.return_value = True
        lines = ["", "\t", " "]
        m_load_file.return_value = "\n".join(lines)
        ret = ssh_util.parse_ssh_config("some real file")
        assert len(lines) == len(ret)
        for line in ret:
            assert "" == line.line

    @pytest.mark.parametrize(
        "file_content, expected_key, expected_value",
        [
            pytest.param("foo bar", "foo", "bar", id="lower-case"),
            pytest.param("Foo Bar", "foo", "Bar", id="upper-case"),
            pytest.param("foo=bar", "foo", "bar", id="lower-case-with-equals"),
            pytest.param("Foo=bar", "foo", "bar", id="upper-case-with-equals"),
        ],
    )
    def test_case_config(
        self,
        m_is_file,
        m_load_file,
        file_content,
        expected_key,
        expected_value,
    ):
        m_is_file.return_value = True
        m_load_file.return_value = file_content
        ret = ssh_util.parse_ssh_config("some real file")
        assert 1 == len(ret)
        assert expected_key == ret[0].key
        assert expected_value == ret[0].value

    def test_duplicated_keys(self, m_is_file, m_load_file):
        file_content = [
            "HostCertificate /data/ssh/ssh_host_rsa_cert",
            "HostCertificate /data/ssh/ssh_host_ed25519_cert",
        ]
        m_is_file.return_value = True
        m_load_file.return_value = "\n".join(file_content)
        ret = ssh_util.parse_ssh_config("some real file")
        assert len(file_content) == len(ret)
        for i in range(len(file_content)):
            assert file_content[i] == ret[i].line


class TestUpdateSshConfigLines:
    """Test the update_ssh_config_lines method."""

    exlines = [
        "#PasswordAuthentication yes",
        "UsePAM yes",
        "# Comment line",
        "AcceptEnv LANG LC_*",
        "X11Forwarding no",
    ]
    pwauth = "PasswordAuthentication"

    def check_line(self, line, opt, val):
        assert line.key == opt.lower()
        assert line.value == val
        assert opt in str(line)
        assert val in str(line)

    @pytest.mark.parametrize(
        "key, value",
        [
            pytest.param("MyKey", "MyVal", id="new_option_added"),
            pytest.param(
                pwauth, "no", id="commented_out_not_updated_but_appended"
            ),
        ],
    )
    def test_update_ssh_config_lines(self, key, value):
        lines = ssh_util.parse_ssh_config_lines(list(self.exlines))
        result = ssh_util.update_ssh_config_lines(lines, {key: value})
        assert [key] == result
        self.check_line(lines[-1], key, value)

    def test_option_without_value(self):
        """Implementation only accepts key-value pairs."""
        extended_exlines = self.exlines.copy()
        denyusers_opt = "DenyUsers"
        extended_exlines.append(denyusers_opt)
        lines = ssh_util.parse_ssh_config_lines(list(extended_exlines))
        assert denyusers_opt not in str(lines)

    def test_single_option_updated(self):
        """A single update should have change made and line updated."""
        opt, val = ("UsePAM", "no")
        lines = ssh_util.parse_ssh_config_lines(list(self.exlines))
        result = ssh_util.update_ssh_config_lines(lines, {opt: val})
        assert [opt] == result
        self.check_line(lines[1], opt, val)

    def test_multiple_updates_with_add(self):
        """Verify multiple updates some added some changed, some not."""
        updates = {
            "UsePAM": "no",
            "X11Forwarding": "no",
            "NewOpt": "newval",
            "AcceptEnv": "LANG ADD LC_*",
        }
        lines = ssh_util.parse_ssh_config_lines(list(self.exlines))
        result = ssh_util.update_ssh_config_lines(lines, updates)
        assert set(["UsePAM", "NewOpt", "AcceptEnv"]) == set(result)
        self.check_line(lines[3], "AcceptEnv", updates["AcceptEnv"])

    def test_return_empty_if_no_changes(self):
        """If there are no changes, then return should be empty list."""
        updates = {"UsePAM": "yes"}
        lines = ssh_util.parse_ssh_config_lines(list(self.exlines))
        result = ssh_util.update_ssh_config_lines(lines, updates)
        assert [] == result
        assert self.exlines == [str(line) for line in lines]

    def test_keycase_not_modified(self):
        """Original case of key should not be changed on update.
        This behavior is to keep original config as much intact as can be."""
        updates = {"usepam": "no"}
        lines = ssh_util.parse_ssh_config_lines(list(self.exlines))
        result = ssh_util.update_ssh_config_lines(lines, updates)
        assert ["usepam"] == result
        assert "UsePAM no" == str(lines[1])


class TestUpdateSshConfig:
    cfgdata = "\n".join(["#Option val", "MyKey ORIG_VAL", ""])

    def test_modified(self, tmpdir):
        mycfg = tmpdir.join("ssh_config_1")
        util.write_file(mycfg, self.cfgdata)
        ret = ssh_util.update_ssh_config({"MyKey": "NEW_VAL"}, mycfg)
        assert True is ret
        found = util.load_text_file(mycfg)
        assert self.cfgdata.replace("ORIG_VAL", "NEW_VAL") == found
        # assert there is a newline at end of file (LP: #1677205)
        assert "\n" == found[-1]

    def test_not_modified(self, tmpdir):
        mycfg = tmpdir.join("ssh_config_2")
        util.write_file(mycfg, self.cfgdata)
        with patch("cloudinit.ssh_util.util.write_file") as m_write_file:
            ret = ssh_util.update_ssh_config({"MyKey": "ORIG_VAL"}, mycfg)
        assert False is ret
        assert self.cfgdata == util.load_text_file(mycfg)
        m_write_file.assert_not_called()

    def test_without_include(self, tmpdir):
        mycfg = tmpdir.join("sshd_config")
        cfg = "X Y"
        util.write_file(mycfg, cfg)
        assert ssh_util.update_ssh_config({"key": "value"}, mycfg)
        assert "X Y\nkey value\n" == util.load_text_file(mycfg)
        expected_conf_file = f"{mycfg}.d/50-cloud-init.conf"
        assert not os.path.isfile(expected_conf_file)

    @pytest.mark.parametrize(
        "cfg",
        ["Include {mycfg}.d/*.conf", "Include {mycfg}.d/*.conf # comment"],
    )
    def test_with_include(self, cfg, tmpdir):
        mycfg = tmpdir.join("sshd_config")
        util.write_file(mycfg, cfg.format(mycfg=mycfg))
        assert ssh_util.update_ssh_config({"key": "value"}, mycfg)
        expected_conf_file = f"{mycfg}.d/50-cloud-init.conf"
        assert os.path.isfile(expected_conf_file)
        assert 0o600 == stat.S_IMODE(os.stat(expected_conf_file).st_mode)
        assert "key value\n" == util.load_text_file(expected_conf_file)

    def test_with_commented_include(self, tmpdir):
        mycfg = tmpdir.join("sshd_config")
        cfg = f"# Include {mycfg}.d/*.conf"
        util.write_file(mycfg, cfg)
        assert ssh_util.update_ssh_config({"key": "value"}, mycfg)
        assert f"{cfg}\nkey value\n" == util.load_text_file(mycfg)
        expected_conf_file = f"{mycfg}.d/50-cloud-init.conf"
        assert not os.path.isfile(expected_conf_file)

    def test_with_other_include(self, tmpdir):
        mycfg = tmpdir.join("sshd_config")
        cfg = f"Include other_{mycfg}.d/*.conf"
        util.write_file(mycfg, cfg)
        assert ssh_util.update_ssh_config({"key": "value"}, mycfg)
        assert f"{cfg}\nkey value\n" == util.load_text_file(mycfg)
        expected_conf_file = f"{mycfg}.d/50-cloud-init.conf"
        assert not os.path.isfile(expected_conf_file)
        assert not os.path.isfile(f"other_{mycfg}.d/50-cloud-init.conf")


class TestAppendSshConfig:
    cfgdata = "\n".join(["#Option val", "MyKey ORIG_VAL", ""])

    @mock.patch(M_PATH + "_ensure_cloud_init_ssh_config_file")
    def test_append_ssh_config(self, m_ensure_cloud_init_config_file, tmpdir):
        mycfg = tmpdir.join("ssh_config")
        util.write_file(mycfg, self.cfgdata)
        m_ensure_cloud_init_config_file.return_value = str(mycfg)
        ssh_util.append_ssh_config(
            [("MyKey", "NEW_VAL"), ("MyKey", "NEW_VAL_2")], mycfg
        )
        found = util.load_text_file(mycfg)
        expected_cfg = dedent(
            """\
            #Option val
            MyKey ORIG_VAL
            MyKey NEW_VAL
            MyKey NEW_VAL_2
            """
        )
        assert expected_cfg == found
        # assert there is a newline at end of file (LP: #1677205)
        assert "\n" == found[-1]


class TestBasicAuthorizedKeyParse:
    @pytest.mark.parametrize(
        "value, homedir, username, expected_rendered",
        [
            pytest.param(
                "/opt/%u/keys",
                "/home/bobby",
                "bobby",
                ["/opt/bobby/keys"],
                id="user",
            ),
            pytest.param(
                "/opt/%u",
                "/home/bobby",
                "bobby",
                ["/opt/bobby"],
                id="user_file",
            ),
            pytest.param(
                "/opt/%u/%u",
                "/home/bobby",
                "bobby",
                ["/opt/bobby/bobby"],
                id="user_file_2",
            ),
            pytest.param(
                "/keys/path1 /keys/path2",
                "/home/bobby",
                "bobby",
                ["/keys/path1", "/keys/path2"],
                id="multiple",
            ),
            pytest.param(
                "/keys/path1 /keys/%u",
                "/home/bobby",
                "bobby",
                ["/keys/path1", "/keys/bobby"],
                id="multiple_2",
            ),
            pytest.param(
                ".secret/keys",
                "/home/bobby",
                "bobby",
                ["/home/bobby/.secret/keys"],
                id="relative",
            ),
            pytest.param(
                "%h/.keys",
                "/homedirs/bobby",
                "bobby",
                ["/homedirs/bobby/.keys"],
                id="home",
            ),
            pytest.param(
                "%h/.keys .secret/keys /keys/path1 /opt/%u/keys",
                "/homedirs/bobby",
                "bobby",
                [
                    "/homedirs/bobby/.keys",
                    "/homedirs/bobby/.secret/keys",
                    "/keys/path1",
                    "/opt/bobby/keys",
                ],
                id="all",
            ),
        ],
    )
    def test_render_authorizedkeysfile_paths(
        self, value, homedir, username, expected_rendered
    ):
        assert expected_rendered == ssh_util.render_authorizedkeysfile_paths(
            value, homedir, username
        )


class TestMultipleSshAuthorizedKeysFile:
    def create_fake_users(
        self,
        names,
        mock_permissions,
        m_get_group,
        m_get_owner,
        m_get_permissions,
        m_getpwnam,
        users,
        tmpdir,
    ):
        homes = []

        root = str(tmpdir.join("root"))
        fpw = FakePwEnt(pw_name="root", pw_dir=root)
        users["root"] = fpw

        for name in names:
            home = str(tmpdir.join("home", name))
            fpw = FakePwEnt(pw_name=name, pw_dir=home)
            users[name] = fpw
            homes.append(home)

        m_get_permissions.side_effect = partial(
            mock_get_permissions, mock_permissions
        )
        m_get_owner.side_effect = partial(mock_get_owner, mock_permissions)
        m_get_group.side_effect = partial(mock_get_group, mock_permissions)
        m_getpwnam.side_effect = partial(mock_getpwnam, users)
        return homes

    def create_user_authorized_file(self, home, filename, content_key, keys):
        user_ssh_folder = os.path.join(home, ".ssh")
        # /tmp/home/<user>/.ssh/authorized_keys = content_key
        authorized_keys = str(os.path.join(user_ssh_folder, filename))
        util.write_file(authorized_keys, VALID_CONTENT[content_key])
        keys[authorized_keys] = content_key
        return authorized_keys

    def create_global_authorized_file(
        self, filename, content_key, keys, tmpdir
    ):
        authorized_keys = str(tmpdir.join(filename))
        util.write_file(authorized_keys, VALID_CONTENT[content_key])
        keys[authorized_keys] = content_key
        return authorized_keys

    def create_sshd_config(self, authorized_keys_files, tmpdir):
        sshd_config = str(tmpdir.join("sshd_config"))
        util.write_file(
            sshd_config, "AuthorizedKeysFile " + authorized_keys_files
        )
        return sshd_config

    def execute_and_check(self, user, sshd_config, solution, keys):
        (auth_key_fn, auth_key_entries) = ssh_util.extract_authorized_keys(
            user, sshd_config
        )
        content = ssh_util.update_authorized_keys(auth_key_entries, [])

        assert auth_key_fn == solution
        for path, key in keys.items():
            if path == solution:
                assert VALID_CONTENT[key] in content
            else:
                assert VALID_CONTENT[key] not in content

    @pytest.mark.parametrize("inverted", [False, True])
    @patch("cloudinit.ssh_util.pwd.getpwnam")
    @patch("cloudinit.util.get_permissions")
    @patch("cloudinit.util.get_owner")
    @patch("cloudinit.util.get_group")
    def test_single_user_two_local_files(
        self,
        m_get_group,
        m_get_owner,
        m_get_permissions,
        m_getpwnam,
        inverted,
        tmpdir,
    ):
        user_bobby = "bobby"
        keys = {}
        users = {}
        mock_permissions = {
            tmpdir.join("home", "bobby"): ("bobby", "bobby", 0o700),
            tmpdir.join("home", "bobby", ".ssh"): ("bobby", "bobby", 0o700),
            tmpdir.join("home", "bobby", ".ssh", "user_keys"): (
                "bobby",
                "bobby",
                0o600,
            ),
            tmpdir.join("home", "bobby", ".ssh", "authorized_keys"): (
                "bobby",
                "bobby",
                0o600,
            ),
        }

        homes = self.create_fake_users(
            [user_bobby],
            mock_permissions,
            m_get_group,
            m_get_owner,
            m_get_permissions,
            m_getpwnam,
            users,
            tmpdir,
        )
        home = homes[0]

        # /tmp/home/bobby/.ssh/authorized_keys = rsa
        authorized_keys = self.create_user_authorized_file(
            home, "authorized_keys", "rsa", keys
        )

        # /tmp/home/bobby/.ssh/user_keys = ed25519
        user_keys = self.create_user_authorized_file(
            home, "user_keys", "ed25519", keys
        )

        # /tmp/sshd_config
        if not inverted:
            options = f"{authorized_keys} {user_keys}"
        else:
            options = f"{user_keys} {authorized_keys}"
        sshd_config = self.create_sshd_config(options, tmpdir)

        if not inverted:
            exec_args = (user_bobby, sshd_config, authorized_keys, keys)
        else:
            exec_args = (user_bobby, sshd_config, user_keys, keys)

        self.execute_and_check(*exec_args)

    @pytest.mark.parametrize("inverted", [False, True])
    @patch("cloudinit.ssh_util.pwd.getpwnam")
    @patch("cloudinit.util.get_permissions")
    @patch("cloudinit.util.get_owner")
    @patch("cloudinit.util.get_group")
    def test_single_user_local_global_files(
        self,
        m_get_group,
        m_get_owner,
        m_get_permissions,
        m_getpwnam,
        inverted,
        tmpdir,
    ):
        user_bobby = "bobby"
        keys = {}
        users = {}
        mock_permissions = {
            tmpdir.join("home", "bobby"): ("bobby", "bobby", 0o700),
            tmpdir.join("home", "bobby", ".ssh"): ("bobby", "bobby", 0o700),
            tmpdir.join("home", "bobby", ".ssh", "user_keys"): (
                "bobby",
                "bobby",
                0o600,
            ),
            tmpdir.join("home", "bobby", ".ssh", "authorized_keys"): (
                "bobby",
                "bobby",
                0o600,
            ),
        }

        homes = self.create_fake_users(
            [user_bobby],
            mock_permissions,
            m_get_group,
            m_get_owner,
            m_get_permissions,
            m_getpwnam,
            users,
            tmpdir,
        )
        home = homes[0]

        # /tmp/home/bobby/.ssh/authorized_keys = rsa
        authorized_keys = self.create_user_authorized_file(
            home, "authorized_keys", "rsa", keys
        )

        # /tmp/home/bobby/.ssh/user_keys = ed25519
        user_keys = self.create_user_authorized_file(
            home, "user_keys", "ed25519", keys
        )

        authorized_keys_global = self.create_global_authorized_file(
            "etc/ssh/authorized_keys", "ecdsa", keys, tmpdir
        )

        if not inverted:
            options = f"{authorized_keys_global} {user_keys} {authorized_keys}"
        else:
            options = f"{authorized_keys_global} {authorized_keys} {user_keys}"
        sshd_config = self.create_sshd_config(options, tmpdir)

        if not inverted:
            exec_args = (user_bobby, sshd_config, user_keys, keys)
        else:
            exec_args = (user_bobby, sshd_config, authorized_keys, keys)
        self.execute_and_check(*exec_args)

    @patch("cloudinit.ssh_util.pwd.getpwnam")
    @patch("cloudinit.util.get_permissions")
    @patch("cloudinit.util.get_owner")
    @patch("cloudinit.util.get_group")
    def test_single_user_global_file(
        self, m_get_group, m_get_owner, m_get_permissions, m_getpwnam, tmpdir
    ):
        user_bobby = "bobby"
        keys = {}
        users = {}
        mock_permissions = {
            tmpdir.join("home", "bobby"): ("bobby", "bobby", 0o700),
            tmpdir.join("home", "bobby", ".ssh"): ("bobby", "bobby", 0o700),
            tmpdir.join("home", "bobby", ".ssh", "authorized_keys"): (
                "bobby",
                "bobby",
                0o600,
            ),
        }

        homes = self.create_fake_users(
            [user_bobby],
            mock_permissions,
            m_get_group,
            m_get_owner,
            m_get_permissions,
            m_getpwnam,
            users,
            tmpdir,
        )
        home = homes[0]

        # /tmp/etc/ssh/authorized_keys = rsa
        authorized_keys_global = self.create_global_authorized_file(
            "etc/ssh/authorized_keys", "rsa", keys, tmpdir
        )

        options = "%s" % authorized_keys_global
        sshd_config = self.create_sshd_config(options, tmpdir)

        default = "%s/.ssh/authorized_keys" % home
        self.execute_and_check(user_bobby, sshd_config, default, keys)

    @patch("cloudinit.ssh_util.pwd.getpwnam")
    @patch("cloudinit.util.get_permissions")
    @patch("cloudinit.util.get_owner")
    @patch("cloudinit.util.get_group")
    def test_two_users_local_file_standard(
        self, m_get_group, m_get_owner, m_get_permissions, m_getpwnam, tmpdir
    ):
        keys = {}
        users = {}
        mock_permissions = {
            tmpdir.join("home", "bobby"): ("bobby", "bobby", 0o700),
            tmpdir.join("home", "bobby", ".ssh"): ("bobby", "bobby", 0o700),
            tmpdir.join("home", "bobby", ".ssh", "authorized_keys"): (
                "bobby",
                "bobby",
                0o600,
            ),
            tmpdir.join("home", "suzie"): ("suzie", "suzie", 0o700),
            tmpdir.join("home", "suzie", ".ssh"): ("suzie", "suzie", 0o700),
            tmpdir.join("home", "suzie", ".ssh", "authorized_keys"): (
                "suzie",
                "suzie",
                0o600,
            ),
        }

        user_bobby = "bobby"
        user_suzie = "suzie"
        homes = self.create_fake_users(
            [user_bobby, user_suzie],
            mock_permissions,
            m_get_group,
            m_get_owner,
            m_get_permissions,
            m_getpwnam,
            users,
            tmpdir,
        )
        home_bobby = homes[0]
        home_suzie = homes[1]

        # /tmp/home/bobby/.ssh/authorized_keys = rsa
        authorized_keys = self.create_user_authorized_file(
            home_bobby, "authorized_keys", "rsa", keys
        )

        # /tmp/home/suzie/.ssh/authorized_keys = rsa
        authorized_keys2 = self.create_user_authorized_file(
            home_suzie, "authorized_keys", "ssh-xmss@openssh.com", keys
        )

        options = ".ssh/authorized_keys"
        sshd_config = self.create_sshd_config(options, tmpdir)

        self.execute_and_check(user_bobby, sshd_config, authorized_keys, keys)
        self.execute_and_check(user_suzie, sshd_config, authorized_keys2, keys)

    @patch("cloudinit.ssh_util.pwd.getpwnam")
    @patch("cloudinit.util.get_permissions")
    @patch("cloudinit.util.get_owner")
    @patch("cloudinit.util.get_group")
    def test_two_users_local_file_custom(
        self, m_get_group, m_get_owner, m_get_permissions, m_getpwnam, tmpdir
    ):
        keys = {}
        users = {}
        mock_permissions = {
            tmpdir.join("home", "bobby"): ("bobby", "bobby", 0o700),
            tmpdir.join("home", "bobby", ".ssh"): ("bobby", "bobby", 0o700),
            tmpdir.join("home", "bobby", ".ssh", "authorized_keys2"): (
                "bobby",
                "bobby",
                0o600,
            ),
            tmpdir.join("home", "suzie"): ("suzie", "suzie", 0o700),
            tmpdir.join("home", "suzie", ".ssh"): ("suzie", "suzie", 0o700),
            tmpdir.join("home", "suzie", ".ssh", "authorized_keys2"): (
                "suzie",
                "suzie",
                0o600,
            ),
        }

        user_bobby = "bobby"
        user_suzie = "suzie"
        homes = self.create_fake_users(
            [user_bobby, user_suzie],
            mock_permissions,
            m_get_group,
            m_get_owner,
            m_get_permissions,
            m_getpwnam,
            users,
            tmpdir,
        )
        home_bobby = homes[0]
        home_suzie = homes[1]

        # /tmp/home/bobby/.ssh/authorized_keys2 = rsa
        authorized_keys = self.create_user_authorized_file(
            home_bobby, "authorized_keys2", "rsa", keys
        )

        # /tmp/home/suzie/.ssh/authorized_keys2 = rsa
        authorized_keys2 = self.create_user_authorized_file(
            home_suzie, "authorized_keys2", "ssh-xmss@openssh.com", keys
        )

        options = ".ssh/authorized_keys2"
        sshd_config = self.create_sshd_config(options, tmpdir)

        self.execute_and_check(user_bobby, sshd_config, authorized_keys, keys)
        self.execute_and_check(user_suzie, sshd_config, authorized_keys2, keys)

    @patch("cloudinit.ssh_util.pwd.getpwnam")
    @patch("cloudinit.util.get_permissions")
    @patch("cloudinit.util.get_owner")
    @patch("cloudinit.util.get_group")
    def test_two_users_local_global_files(
        self, m_get_group, m_get_owner, m_get_permissions, m_getpwnam, tmpdir
    ):
        keys = {}
        users = {}
        mock_permissions = {
            tmpdir.join("home", "bobby"): ("bobby", "bobby", 0o700),
            tmpdir.join("home", "bobby", ".ssh"): ("bobby", "bobby", 0o700),
            tmpdir.join("home", "bobby", ".ssh", "authorized_keys2"): (
                "bobby",
                "bobby",
                0o600,
            ),
            tmpdir.join("home", "bobby", ".ssh", "user_keys3"): (
                "bobby",
                "bobby",
                0o600,
            ),
            tmpdir.join("home", "suzie"): ("suzie", "suzie", 0o700),
            tmpdir.join("home", "suzie", ".ssh"): ("suzie", "suzie", 0o700),
            tmpdir.join("home", "suzie", ".ssh", "authorized_keys2"): (
                "suzie",
                "suzie",
                0o600,
            ),
            tmpdir.join("home", "suzie", ".ssh", "user_keys3"): (
                "suzie",
                "suzie",
                0o600,
            ),
        }

        user_bobby = "bobby"
        user_suzie = "suzie"
        homes = self.create_fake_users(
            [user_bobby, user_suzie],
            mock_permissions,
            m_get_group,
            m_get_owner,
            m_get_permissions,
            m_getpwnam,
            users,
            tmpdir,
        )
        home_bobby = homes[0]
        home_suzie = homes[1]

        # /tmp/home/bobby/.ssh/authorized_keys2 = rsa
        self.create_user_authorized_file(
            home_bobby, "authorized_keys2", "rsa", keys
        )
        # /tmp/home/bobby/.ssh/user_keys3 = ed25519
        user_keys = self.create_user_authorized_file(
            home_bobby, "user_keys3", "ed25519", keys
        )

        # /tmp/home/suzie/.ssh/authorized_keys2 = rsa
        authorized_keys2 = self.create_user_authorized_file(
            home_suzie, "authorized_keys2", "ssh-xmss@openssh.com", keys
        )

        # /tmp/etc/ssh/authorized_keys = ecdsa
        authorized_keys_global = self.create_global_authorized_file(
            "etc/ssh/authorized_keys2", "ecdsa", keys, tmpdir
        )

        options = "%s %s %%h/.ssh/authorized_keys2" % (
            authorized_keys_global,
            user_keys,
        )
        sshd_config = self.create_sshd_config(options, tmpdir)

        self.execute_and_check(user_bobby, sshd_config, user_keys, keys)
        self.execute_and_check(user_suzie, sshd_config, authorized_keys2, keys)

    @patch("cloudinit.util.get_user_groups")
    @patch("cloudinit.ssh_util.pwd.getpwnam")
    @patch("cloudinit.util.get_permissions")
    @patch("cloudinit.util.get_owner")
    @patch("cloudinit.util.get_group")
    def test_two_users_local_global_files_badguy(
        self,
        m_get_group,
        m_get_owner,
        m_get_permissions,
        m_getpwnam,
        m_get_user_groups,
        tmpdir,
    ):
        keys = {}
        users = {}
        mock_permissions = {
            tmpdir.join("home", "bobby"): ("bobby", "bobby", 0o700),
            tmpdir.join("home", "bobby", ".ssh"): ("bobby", "bobby", 0o700),
            tmpdir.join("home", "bobby", ".ssh", "authorized_keys2"): (
                "bobby",
                "bobby",
                0o600,
            ),
            tmpdir.join("home", "bobby", ".ssh", "user_keys3"): (
                "bobby",
                "bobby",
                0o600,
            ),
            tmpdir.join("home", "badguy"): ("root", "root", 0o755),
            tmpdir.join("home", "badguy", "home"): ("root", "root", 0o755),
            tmpdir.join("home", "badguy", "home", "bobby"): (
                "root",
                "root",
                0o655,
            ),
        }

        user_bobby = "bobby"
        user_badguy = "badguy"
        home_bobby, *_ = self.create_fake_users(
            [user_bobby, user_badguy],
            mock_permissions,
            m_get_group,
            m_get_owner,
            m_get_permissions,
            m_getpwnam,
            users,
            tmpdir,
        )
        m_get_user_groups.side_effect = mock_get_user_groups

        # /tmp/home/bobby/.ssh/authorized_keys2 = rsa
        authorized_keys = self.create_user_authorized_file(
            home_bobby, "authorized_keys2", "rsa", keys
        )
        # /tmp/home/bobby/.ssh/user_keys3 = ecdsa
        user_keys = self.create_user_authorized_file(
            home_bobby, "user_keys3", "ecdsa", keys
        )

        # /tmp/home/badguy/home/bobby = ""
        authorized_keys2 = str(tmpdir.join("home", "badguy", "home", "bobby"))
        util.write_file(authorized_keys2, "")

        # /tmp/etc/ssh/authorized_keys = ecdsa
        authorized_keys_global = self.create_global_authorized_file(
            "etc/ssh/authorized_keys2", "ecdsa", keys, tmpdir
        )

        # /tmp/sshd_config
        options = "%s %%h/.ssh/authorized_keys2 %s %s" % (
            authorized_keys2,
            authorized_keys_global,
            user_keys,
        )
        sshd_config = self.create_sshd_config(options, tmpdir)

        self.execute_and_check(user_bobby, sshd_config, authorized_keys, keys)
        self.execute_and_check(
            user_badguy, sshd_config, authorized_keys2, keys
        )

    @patch("cloudinit.util.get_user_groups")
    @patch("cloudinit.ssh_util.pwd.getpwnam")
    @patch("cloudinit.util.get_permissions")
    @patch("cloudinit.util.get_owner")
    @patch("cloudinit.util.get_group")
    def test_two_users_unaccessible_file(
        self,
        m_get_group,
        m_get_owner,
        m_get_permissions,
        m_getpwnam,
        m_get_user_groups,
        tmpdir,
    ):
        keys = {}
        users = {}
        mock_permissions = {
            tmpdir.join("home", "bobby"): ("bobby", "bobby", 0o700),
            tmpdir.join("home", "bobby", ".ssh"): ("bobby", "bobby", 0o700),
            tmpdir.join("home", "bobby", ".ssh", "authorized_keys"): (
                "bobby",
                "bobby",
                0o600,
            ),
            tmpdir.join("etc"): ("root", "root", 0o755),
            tmpdir.join("etc", "ssh"): ("root", "root", 0o755),
            tmpdir.join("etc", "ssh", "userkeys"): ("root", "root", 0o700),
            tmpdir.join("etc", "ssh", "userkeys", "bobby"): (
                "bobby",
                "bobby",
                0o600,
            ),
            tmpdir.join("etc", "ssh", "userkeys", "badguy"): (
                "badguy",
                "badguy",
                0o600,
            ),
            tmpdir.join("home", "badguy"): ("badguy", "badguy", 0o700),
            tmpdir.join("home", "badguy", ".ssh"): ("badguy", "badguy", 0o700),
            tmpdir.join("home", "badguy", ".ssh", "authorized_keys"): (
                "badguy",
                "badguy",
                0o600,
            ),
        }

        user_bobby = "bobby"
        user_badguy = "badguy"
        homes = self.create_fake_users(
            [user_bobby, user_badguy],
            mock_permissions,
            m_get_group,
            m_get_owner,
            m_get_permissions,
            m_getpwnam,
            users,
            tmpdir,
        )
        m_get_user_groups.side_effect = mock_get_user_groups
        home_bobby = homes[0]
        home_badguy = homes[1]

        # /tmp/home/bobby/.ssh/authorized_keys = rsa
        authorized_keys = self.create_user_authorized_file(
            home_bobby, "authorized_keys", "rsa", keys
        )
        # /tmp/etc/ssh/userkeys/bobby = ecdsa
        # assume here that we can bypass userkeys, despite permissions
        self.create_global_authorized_file(
            "etc/ssh/userkeys/bobby", "ecdsa", keys, tmpdir
        )

        # /tmp/home/badguy/.ssh/authorized_keys = ssh-xmss@openssh.com
        authorized_keys2 = self.create_user_authorized_file(
            home_badguy, "authorized_keys", "ssh-xmss@openssh.com", keys
        )

        # /tmp/etc/ssh/userkeys/badguy = ecdsa
        self.create_global_authorized_file(
            "etc/ssh/userkeys/badguy", "ecdsa", keys, tmpdir
        )

        # /tmp/sshd_config
        options = str(
            tmpdir.join("etc", "ssh", "userkeys", "%u .ssh", "authorized_keys")
        )
        sshd_config = self.create_sshd_config(options, tmpdir)

        self.execute_and_check(user_bobby, sshd_config, authorized_keys, keys)
        self.execute_and_check(
            user_badguy, sshd_config, authorized_keys2, keys
        )

    @patch("cloudinit.util.get_user_groups")
    @patch("cloudinit.ssh_util.pwd.getpwnam")
    @patch("cloudinit.util.get_permissions")
    @patch("cloudinit.util.get_owner")
    @patch("cloudinit.util.get_group")
    def test_two_users_accessible_file(
        self,
        m_get_group,
        m_get_owner,
        m_get_permissions,
        m_getpwnam,
        m_get_user_groups,
        tmpdir,
    ):
        keys = {}
        users = {}
        mock_permissions = {
            tmpdir.join("home", "bobby"): ("bobby", "bobby", 0o700),
            tmpdir.join("home", "bobby", ".ssh"): ("bobby", "bobby", 0o700),
            tmpdir.join("home", "bobby", ".ssh", "authorized_keys"): (
                "bobby",
                "bobby",
                0o600,
            ),
            tmpdir.join("etc"): ("root", "root", 0o755),
            tmpdir.join("etc", "ssh"): ("root", "root", 0o755),
            tmpdir.join("etc", "ssh", "userkeys"): ("root", "root", 0o755),
            tmpdir.join("etc", "ssh", "userkeys", "bobby"): (
                "bobby",
                "bobby",
                0o600,
            ),
            tmpdir.join("etc", "ssh", "userkeys", "badguy"): (
                "badguy",
                "badguy",
                0o600,
            ),
            tmpdir.join("home", "badguy"): ("badguy", "badguy", 0o700),
            tmpdir.join("home", "badguy", ".ssh"): ("badguy", "badguy", 0o700),
            tmpdir.join("home", "badguy", ".ssh", "authorized_keys"): (
                "badguy",
                "badguy",
                0o600,
            ),
        }

        user_bobby = "bobby"
        user_badguy = "badguy"
        homes = self.create_fake_users(
            [user_bobby, user_badguy],
            mock_permissions,
            m_get_group,
            m_get_owner,
            m_get_permissions,
            m_getpwnam,
            users,
            tmpdir,
        )
        m_get_user_groups.side_effect = mock_get_user_groups
        home_bobby = homes[0]
        home_badguy = homes[1]

        # /tmp/home/bobby/.ssh/authorized_keys = rsa
        self.create_user_authorized_file(
            home_bobby, "authorized_keys", "rsa", keys
        )
        # /tmp/etc/ssh/userkeys/bobby = ed25519
        # assume here that we can bypass userkeys, despite permissions
        authorized_keys = self.create_global_authorized_file(
            "etc/ssh/userkeys/bobby", "ed25519", keys, tmpdir
        )

        # /tmp/home/badguy/.ssh/authorized_keys = ssh-xmss@openssh.com
        self.create_user_authorized_file(
            home_badguy, "authorized_keys", "ssh-xmss@openssh.com", keys
        )

        # /tmp/etc/ssh/userkeys/badguy = ecdsa
        authorized_keys2 = self.create_global_authorized_file(
            "etc/ssh/userkeys/badguy", "ecdsa", keys, tmpdir
        )

        # /tmp/sshd_config
        options = str(
            tmpdir.join("etc", "ssh", "userkeys", "%u .ssh", "authorized_keys")
        )
        sshd_config = self.create_sshd_config(options, tmpdir)

        self.execute_and_check(user_bobby, sshd_config, authorized_keys, keys)
        self.execute_and_check(
            user_badguy, sshd_config, authorized_keys2, keys
        )

    @pytest.mark.parametrize("inverted", [False, True])
    @patch("cloudinit.util.get_user_groups")
    @patch("cloudinit.ssh_util.pwd.getpwnam")
    @patch("cloudinit.util.get_permissions")
    @patch("cloudinit.util.get_owner")
    @patch("cloudinit.util.get_group")
    def test_two_users_hardcoded_single_user_file(
        self,
        m_get_group,
        m_get_owner,
        m_get_permissions,
        m_getpwnam,
        m_get_user_groups,
        inverted,
        tmpdir,
    ):
        keys = {}
        users = {}
        mock_permissions = {
            tmpdir.join("home", "bobby"): ("bobby", "bobby", 0o700),
            tmpdir.join("home", "bobby", ".ssh"): ("bobby", "bobby", 0o700),
            tmpdir.join("home", "bobby", ".ssh", "authorized_keys"): (
                "bobby",
                "bobby",
                0o600,
            ),
            tmpdir.join("home", "suzie"): ("suzie", "suzie", 0o700),
            tmpdir.join("home", "suzie", ".ssh"): ("suzie", "suzie", 0o700),
            tmpdir.join("home", "suzie", ".ssh", "authorized_keys"): (
                "suzie",
                "suzie",
                0o600,
            ),
        }

        user_bobby = "bobby"
        user_suzie = "suzie"
        homes = self.create_fake_users(
            [user_bobby, user_suzie],
            mock_permissions,
            m_get_group,
            m_get_owner,
            m_get_permissions,
            m_getpwnam,
            users,
            tmpdir,
        )
        home_bobby = homes[0]
        home_suzie = homes[1]
        m_get_user_groups.side_effect = mock_get_user_groups

        # /tmp/home/bobby/.ssh/authorized_keys = rsa
        authorized_keys = self.create_user_authorized_file(
            home_bobby, "authorized_keys", "rsa", keys
        )

        # /tmp/home/suzie/.ssh/authorized_keys = ssh-xmss@openssh.com
        authorized_keys2 = self.create_user_authorized_file(
            home_suzie, "authorized_keys", "ssh-xmss@openssh.com", keys
        )

        # /tmp/sshd_config
        if not inverted:
            expected_keys = authorized_keys
        else:
            expected_keys = authorized_keys2
        options = "%s" % (expected_keys)
        sshd_config = self.create_sshd_config(options, tmpdir)

        if not inverted:
            expected_bobby = expected_keys
            expected_suzie = "%s/.ssh/authorized_keys" % home_suzie
        else:
            expected_bobby = "%s/.ssh/authorized_keys" % home_bobby
            expected_suzie = expected_keys
        self.execute_and_check(user_bobby, sshd_config, expected_bobby, keys)
        self.execute_and_check(user_suzie, sshd_config, expected_suzie, keys)

    @patch("cloudinit.util.get_user_groups")
    @patch("cloudinit.ssh_util.pwd.getpwnam")
    @patch("cloudinit.util.get_permissions")
    @patch("cloudinit.util.get_owner")
    @patch("cloudinit.util.get_group")
    def test_two_users_hardcoded_user_files(
        self,
        m_get_group,
        m_get_owner,
        m_get_permissions,
        m_getpwnam,
        m_get_user_groups,
        tmpdir,
    ):
        keys = {}
        users = {}
        mock_permissions = {
            tmpdir.join("home", "bobby"): ("bobby", "bobby", 0o700),
            tmpdir.join("home", "bobby", ".ssh"): ("bobby", "bobby", 0o700),
            tmpdir.join("home", "bobby", ".ssh", "authorized_keys"): (
                "bobby",
                "bobby",
                0o600,
            ),
            tmpdir.join("home", "suzie"): ("suzie", "suzie", 0o700),
            tmpdir.join("home", "suzie", ".ssh"): ("suzie", "suzie", 0o700),
            tmpdir.join("home", "suzie", ".ssh", "authorized_keys"): (
                "suzie",
                "suzie",
                0o600,
            ),
        }

        user_bobby = "bobby"
        user_suzie = "suzie"
        homes = self.create_fake_users(
            [user_bobby, user_suzie],
            mock_permissions,
            m_get_group,
            m_get_owner,
            m_get_permissions,
            m_getpwnam,
            users,
            tmpdir,
        )
        home_bobby = homes[0]
        home_suzie = homes[1]
        m_get_user_groups.side_effect = mock_get_user_groups

        # /tmp/home/bobby/.ssh/authorized_keys = rsa
        authorized_keys = self.create_user_authorized_file(
            home_bobby, "authorized_keys", "rsa", keys
        )

        # /tmp/home/suzie/.ssh/authorized_keys = ssh-xmss@openssh.com
        authorized_keys2 = self.create_user_authorized_file(
            home_suzie, "authorized_keys", "ssh-xmss@openssh.com", keys
        )

        # /tmp/etc/ssh/authorized_keys = ecdsa
        authorized_keys_global = self.create_global_authorized_file(
            "etc/ssh/authorized_keys", "ecdsa", keys, tmpdir
        )

        # /tmp/sshd_config
        options = "%s %s %s" % (
            authorized_keys_global,
            authorized_keys,
            authorized_keys2,
        )
        sshd_config = self.create_sshd_config(options, tmpdir)

        self.execute_and_check(user_bobby, sshd_config, authorized_keys, keys)
        self.execute_and_check(user_suzie, sshd_config, authorized_keys2, keys)



"""Integration test for the ntp module's ntp functionality.

This test specifies the use of the `ntp` NTP client, and ensures that the given
NTP servers are configured as expected.

(This is ported from ``tests/cloud_tests/testcases/modules/ntp_servers.yaml``,
``tests/cloud_tests/testcases/modules/ntp_pools.yaml``,
and ``tests/cloud_tests/testcases/modules/ntp_chrony.yaml``)
"""
import re

import pytest
import yaml

from tests.integration_tests.instances import IntegrationInstance

USER_DATA = """\
#cloud-config
ntp:
  ntp_client: ntp
  servers:
      - 172.16.15.14
      - 172.16.17.18
  pools:
      - 0.cloud-init.mypool
      - 1.cloud-init.mypool
      - 172.16.15.15
"""

EXPECTED_SERVERS = yaml.safe_load(USER_DATA)["ntp"]["servers"]
EXPECTED_POOLS = yaml.safe_load(USER_DATA)["ntp"]["pools"]


@pytest.mark.user_data(USER_DATA)
class TestNtpServers:
    def test_ntp_installed(self, class_client: IntegrationInstance):
        """Test that `ntpd --version` succeeds, indicating installation."""
        assert class_client.execute("ntpd --version").ok

    def test_dist_config_file_is_empty(
        self, class_client: IntegrationInstance
    ):
        """Test that the distributed config file is empty.

        (This test is skipped on all currently supported Ubuntu releases, so
        may not actually be needed any longer.)
        """
        if class_client.execute("test -e /etc/ntp.conf.dist").failed:
            pytest.skip("/etc/ntp.conf.dist does not exist")
        dist_file = class_client.read_from_file("/etc/ntp.conf.dist")
        assert 0 == len(dist_file.strip().splitlines())

    def test_ntp_entries(self, class_client: IntegrationInstance):
        ntp_conf = class_client.read_from_file("/etc/ntp.conf")
        for expected_server in EXPECTED_SERVERS:
            assert re.search(
                r"^server {} iburst".format(expected_server),
                ntp_conf,
                re.MULTILINE,
            )
        for expected_pool in EXPECTED_POOLS:
            assert re.search(
                r"^pool {} iburst".format(expected_pool),
                ntp_conf,
                re.MULTILINE,
            )

    def test_ntpq_servers(self, class_client: IntegrationInstance):
        result = class_client.execute("ntpq -p -w -n")
        assert result.ok
        for expected_server_or_pool in [*EXPECTED_SERVERS, *EXPECTED_POOLS]:
            assert expected_server_or_pool in result.stdout


CHRONY_DATA = """\
#cloud-config
ntp:
  enabled: true
  ntp_client: chrony
  servers:
      - 172.16.15.14
"""


@pytest.mark.user_data(CHRONY_DATA)
def test_chrony(client: IntegrationInstance):
    if client.execute("test -f /etc/chrony.conf").ok:
        chrony_conf = "/etc/chrony.conf"
    else:
        chrony_conf = "/etc/chrony/chrony.conf"
    contents = client.read_from_file(chrony_conf)
    assert "server 172.16.15.14" in contents


TIMESYNCD_DATA = """\
#cloud-config
ntp:
  enabled: true
  ntp_client: systemd-timesyncd
  servers:
      - 172.16.15.14
"""


@pytest.mark.user_data(TIMESYNCD_DATA)
def test_timesyncd(client: IntegrationInstance):
    contents = client.read_from_file(
        "/etc/systemd/timesyncd.conf.d/cloud-init.conf"
    )
    assert "NTP=172.16.15.14" in contents


EMPTY_NTP = """\
#cloud-config
ntp:
  ntp_client: ntp
  pools: []
  servers: []
"""


@pytest.mark.user_data(EMPTY_NTP)
def test_empty_ntp(client: IntegrationInstance):
    assert client.execute("ntpd --version").ok
    assert client.execute("test -f /etc/ntp.conf.dist").failed
    assert "pool.ntp.org iburst" in client.execute(
        'grep -v "^#" /etc/ntp.conf'
    )



import re
from typing import Any, Dict, Optional, Union

from cloudinit import dmi

# dmi data can override
DMI_OVERRIDE_MAP = {
    "als": "allow_local_stage",
    "ais": "allow_init_stage",
    "dhcp": "allow_dhcp",
    "v4": "allow_ipv4",
    "v6": "allow_ipv6",
    "pmp": "preferred_mac_prefixes",
}


def get_dmi_config() -> Dict[str, Union[bool, str]]:
    """
    Parses flags from dmi data and updates self.ds_cfg accordingly
    """
    dmi_flags = dmi.read_dmi_data("baseboard-serial-number")
    ret: Dict[str, Any] = {}

    if not dmi_flags:
        return ret

    # parse the value into individual flags, then set them in our config
    # based on the short name lookup table
    for key, value, _ in re.findall(r"([a-z0-9]+)=(.*?)(;|$)", dmi_flags):
        if key in DMI_OVERRIDE_MAP:
            if value in "01":
                value = bool(int(value))
            elif key == "pmp":
                value = value.split(",")
            ret[DMI_OVERRIDE_MAP[key]] = value

    return ret


def is_on_akamai() -> bool:
    """
    Reads the BIOS vendor from dmi data to determine if we are running in the
    Akamai Connected Cloud.
    """
    vendor = dmi.read_dmi_data("system-manufacturer")
    return vendor in ("Linode", "Akamai")


def get_local_instance_id() -> Optional[str]:
    """
    Returns the instance id read from dmi data without requiring the metadata
    service to be reachable
    """
    return dmi.read_dmi_data("system-serial-number")



# Copyright (C) 2011 Canonical Ltd.
# Copyright (C) 2012 Hewlett-Packard Development Company, L.P.
#
# Author: Scott Moser <scott.moser@canonical.com>
# Author: Juerg Haefliger <juerg.haefliger@hp.com>
#
# This file is part of cloud-init. See LICENSE file for license information.
"""Final Message: Output final message when cloud-init has finished"""

import logging
from textwrap import dedent

from cloudinit import templater, util, version
from cloudinit.cloud import Cloud
from cloudinit.config import Config
from cloudinit.config.schema import MetaSchema, get_meta_doc
from cloudinit.distros import ALL_DISTROS
from cloudinit.settings import PER_ALWAYS

MODULE_DESCRIPTION = """\
This module configures the final message that cloud-init writes. The message is
specified as a jinja template with the following variables set:

    - ``version``: cloud-init version
    - ``timestamp``: time at cloud-init finish
    - ``datasource``: cloud-init data source
    - ``uptime``: system uptime

This message is written to the cloud-init log (usually /var/log/cloud-init.log)
as well as stderr (which usually redirects to /var/log/cloud-init-output.log).

Upon exit, this module writes the system uptime, timestamp, and cloud-init
version to ``/var/lib/cloud/instance/boot-finished`` independent of any
user data specified for this module.
"""
frequency = PER_ALWAYS
meta: MetaSchema = {
    "id": "cc_final_message",
    "name": "Final Message",
    "title": "Output final message when cloud-init has finished",
    "description": MODULE_DESCRIPTION,
    "distros": [ALL_DISTROS],
    "frequency": frequency,
    "examples": [
        dedent(
            """\
            final_message: |
              cloud-init has finished
              version: $version
              timestamp: $timestamp
              datasource: $datasource
              uptime: $uptime
            """
        )
    ],
    "activate_by_schema_keys": [],
}

LOG = logging.getLogger(__name__)
__doc__ = get_meta_doc(meta)

# Jinja formatted default message
FINAL_MESSAGE_DEF = (
    "## template: jinja\n"
    "Cloud-init v. {{version}} finished at {{timestamp}}."
    " Datasource {{datasource}}.  Up {{uptime}} seconds"
)


def handle(name: str, cfg: Config, cloud: Cloud, args: list) -> None:

    msg_in = ""
    if len(args) != 0:
        msg_in = str(args[0])
    else:
        msg_in = util.get_cfg_option_str(cfg, "final_message", "")

    msg_in = msg_in.strip()
    if not msg_in:
        msg_in = FINAL_MESSAGE_DEF

    uptime = util.uptime()
    ts = util.time_rfc2822()
    cver = version.version_string()
    try:
        subs = {
            "uptime": uptime,
            "timestamp": ts,
            "version": cver,
            "datasource": str(cloud.datasource),
        }
        subs.update(dict([(k.upper(), v) for k, v in subs.items()]))
        util.multi_log(
            "%s\n" % (templater.render_string(msg_in, subs)),
            console=False,
            stderr=True,
            log=LOG,
        )
    except templater.JinjaSyntaxParsingException as e:
        util.logexc(
            LOG, "Failed to render templated final message: %s", str(e)
        )
    except Exception:
        util.logexc(LOG, "Failed to render final message template")

    boot_fin_fn = cloud.paths.boot_finished
    try:
        contents = "%s - %s - v. %s\n" % (uptime, ts, cver)
        util.write_file(boot_fin_fn, contents, ensure_dir_exists=False)
    except Exception:
        util.logexc(LOG, "Failed to write boot finished file %s", boot_fin_fn)

    if cloud.datasource.is_disconnected:
        LOG.warning("Used fallback datasource")



# Copyright (C) 2012 Canonical Ltd.
# Copyright (C) 2012 Hewlett-Packard Development Company, L.P.
# Copyright (C) 2012 Yahoo! Inc.
#
# Author: Scott Moser <scott.moser@canonical.com>
# Author: Juerg Haefliger <juerg.haefliger@hp.com>
# Author: Joshua Harlow <harlowja@yahoo-inc.com>
#
# This file is part of cloud-init. See LICENSE file for license information.

# Set and read for determining the cloud config file location
CFG_ENV_NAME = "CLOUD_CFG"

# This is expected to be a yaml formatted file
CLOUD_CONFIG = "/etc/cloud/cloud.cfg"

CLEAN_RUNPARTS_DIR = "/etc/cloud/clean.d"

RUN_CLOUD_CONFIG = "/run/cloud-init/cloud.cfg"

# What u get if no config is provided
CFG_BUILTIN = {
    "datasource_list": [
        "NoCloud",
        "ConfigDrive",
        "LXD",
        "OpenNebula",
        "DigitalOcean",
        "Azure",
        "AltCloud",
        "OVF",
        "MAAS",
        "GCE",
        "OpenStack",
        "AliYun",
        "Vultr",
        "Ec2",
        "CloudSigma",
        "CloudStack",
        "SmartOS",
        "Bigstep",
        "Scaleway",
        "Hetzner",
        "IBMCloud",
        "Oracle",
        "Exoscale",
        "RbxCloud",
        "UpCloud",
        "VMware",
        "NWCS",
        "Akamai",
        "WSL",
        # At the end to act as a 'catch' when none of the above work...
        "None",
    ],
    "def_log_file": "/var/log/cloud-init.log",
    "log_cfgs": [],
    "syslog_fix_perms": ["syslog:adm", "root:adm", "root:wheel", "root:root"],
    "system_info": {
        "paths": {
            "cloud_dir": "/var/lib/cloud",
            "templates_dir": "/etc/cloud/templates/",
        },
        "distro": "ubuntu",
        "network": {"renderers": None},
    },
    "vendor_data": {"enabled": True, "prefix": []},
    "vendor_data2": {"enabled": True, "prefix": []},
}

# Valid frequencies of handlers/modules
PER_INSTANCE = "once-per-instance"
PER_ALWAYS = "always"
PER_ONCE = "once"

# Used to sanity check incoming handlers/modules frequencies
FREQUENCIES = [PER_INSTANCE, PER_ALWAYS, PER_ONCE]

HOTPLUG_ENABLED_FILE = "/var/lib/cloud/hotplug.enabled"



# This file is part of cloud-init. See LICENSE file for license information.
from unittest import mock

from cloudinit import cloud, distros, helpers
from cloudinit.net.dhcp import IscDhclient
from cloudinit.sources import DataSource, DataSourceHostname
from cloudinit.sources.DataSourceNone import DataSourceNone


def get_cloud(
    distro=None, paths=None, sys_cfg=None, metadata=None, mocked_distro=False
):
    """Obtain a "cloud" that can be used for testing.

    Modules take a 'cloud' parameter to call into things that are
    datasource/distro specific. In most cases, the specifics of this cloud
    implementation aren't needed to test the module, so provide a fake
    datasource/distro with stubbed calls to methods that may attempt to
    read/write files or shell out. If a specific distro is needed, it can
    be passed in as the distro parameter.
    """
    paths = paths or helpers.Paths({})
    sys_cfg = sys_cfg or {}
    cls = distros.fetch(distro) if distro else MockDistro
    # *BSD calls platform.system to determine osfamilies
    osfamily = distro.lower() if distro else "ubuntu"
    with mock.patch("platform.system", return_value=osfamily):
        mydist = cls(distro, sys_cfg, paths)
    if mocked_distro:
        mydist = mock.MagicMock(wraps=mydist)
    myds = DataSourceTesting(sys_cfg, mydist, paths)
    if metadata:
        myds.metadata.update(metadata)
    if paths:
        paths.datasource = myds
    return cloud.Cloud(
        myds, paths, sys_cfg, mydist, runners=helpers.Runners(paths)
    )


def abstract_to_concrete(abclass):
    """Takes an abstract class and returns a concrete version of it."""

    class concreteCls(abclass):
        pass

    concreteCls.__abstractmethods__ = frozenset()
    return type("DummyConcrete" + abclass.__name__, (concreteCls,), {})


class DataSourceTesting(DataSourceNone):
    def get_hostname(self, fqdn=False, resolve_ip=False, metadata_only=False):
        return DataSourceHostname("hostname", False)

    def persist_instance_data(self):
        return True

    @property
    def cloud_name(self):
        return "testing"


class MockDistro(distros.Distro):
    # MockDistro is here to test base Distro class implementations
    def __init__(self, name="testingdistro", cfg=None, paths=None):
        self._client = None
        if not cfg:
            cfg = {}
        if not paths:
            paths = {}
        super(MockDistro, self).__init__(name, cfg, paths)

    @property
    def dhcp_client(self):
        if not self._client:
            with mock.patch(
                "cloudinit.net.dhcp.subp.which", return_value=True
            ):
                self._client = IscDhclient()
        return self._client

    def install_packages(self, pkglist):
        pass

    def set_hostname(self, hostname, fqdn=None):
        pass

    @staticmethod
    def uses_systemd():
        return True

    @staticmethod
    def get_proc_ppid(_):
        return 1

    @staticmethod
    def get_proc_pgid(_):
        return 99999

    def get_primary_arch(self):
        return "i386"

    def get_package_mirror_info(self, arch=None, data_source=None):
        pass

    def generate_fallback_config(self):
        return {}

    def apply_network_config(self, netconfig, bring_up=False) -> bool:
        return False

    def apply_locale(self, locale, out_fn=None):
        pass

    def set_timezone(self, tz):
        pass

    def _read_hostname(self, filename, default=None):
        raise NotImplementedError()

    def _write_hostname(self, hostname, filename):
        raise NotImplementedError()

    def _read_system_hostname(self):
        raise NotImplementedError()

    def update_hostname(self, hostname, fqdn, prev_hostname_fn):
        pass

    def update_etc_hosts(self, hostname, fqdn):
        pass

    def add_user(self, name, **kwargs):
        pass

    def add_snap_user(self, name, **kwargs):
        return "snap_user"

    def create_user(self, name, **kwargs):
        return True

    def lock_passwd(self, name):
        pass

    def expire_passwd(self, user):
        pass

    def set_passwd(self, user, passwd, hashed=False):
        return True

    def ensure_sudo_dir(self, path, sudo_base="/etc/sudoers"):
        pass

    def write_sudo_rules(self, user, rules, sudo_file=None):
        pass

    def create_group(self, name, members=None):
        pass

    def shutdown_command(self, *, mode, delay, message):
        pass

    def package_command(self, command, args=None, pkgs=None):
        pass

    def update_package_sources(self):
        return (True, "yay")

    def do_as(self, command, args=None, **kwargs):
        return ("stdout", "stderr")


TEST_INSTANCE_ID = "i-testing"


class FakeDataSource(DataSource):
    def __init__(
        self,
        userdata=None,
        vendordata=None,
        vendordata2=None,
        network_config="",
        paths=None,
    ):
        DataSource.__init__(self, {}, None, paths=paths)
        self.metadata = {"instance-id": TEST_INSTANCE_ID}
        self.userdata_raw = userdata
        self.vendordata_raw = vendordata
        self.vendordata2_raw = vendordata2
        self._network_config = None
        if network_config:  # Permit for None value to setup attribute
            self._network_config = network_config

    @property
    def network_config(self):
        return self._network_config

    def _get_data(self):
        return True



from typing import Any, Dict

import pytest

from cloudinit.sources.helpers.akamai import get_dmi_config, is_on_akamai
from tests.unittests.helpers import mock


class TestAkamaiHelper:
    """
    Test for the Akamai helper functions
    """

    @pytest.mark.parametrize(
        "dmi_config,expected_result",
        (
            # no content
            ("", {}),
            # stage control flags
            (
                "als=0;ais=0",
                {"allow_local_stage": False, "allow_init_stage": False},
            ),
            # networking flags
            (
                "v4=0;v6=0;dhcp=0",
                {
                    "allow_ipv4": False,
                    "allow_ipv6": False,
                    "allow_dhcp": False,
                },
            ),
            # interface prefix overrides
            ("pmp=ab:cd:", {"preferred_mac_prefixes": ["ab:cd:"]}),
            (
                "pmp=ab:cd:,12:34:",
                {"preferred_mac_prefixes": ["ab:cd:", "12:34:"]},
            ),
            # unrecognized data
            ("unrecognized", {}),
            (",;", {}),
        ),
    )
    @mock.patch("cloudinit.dmi.read_dmi_data")
    def test_get_dmi_config(
        self, read_dmi_data, dmi_config: str, expected_result: Dict[str, Any]
    ):
        """
        Tests that dmi-based configuration overrides parse as expected
        """
        read_dmi_data.return_value = dmi_config
        result = get_dmi_config()

        assert (
            result == expected_result
        ), f"Unexpected result parsing dmi config '{dmi_config}'!"
        assert [
            mock.call("baseboard-serial-number")
        ] == read_dmi_data.call_args_list

    @pytest.mark.parametrize(
        "dmi_data,expected_result",
        (
            ("Akamai", True),
            ("Linode", True),
            ("GCE", False),
            ("EC2", False),
            ("", False),
        ),
    )
    @mock.patch("cloudinit.dmi.read_dmi_data")
    def test_is_on_akamai(
        self, read_dmi_data, dmi_data: str, expected_result: bool
    ):
        """
        Tests that is_on_akamai correctly detects if we are on Akama's platform
        """
        read_dmi_data.return_value = dmi_data
        result = is_on_akamai()

        assert (
            result == expected_result
        ), f"Unexpected result checking if '{dmi_data}' is Akamai"
        assert [
            mock.call("system-manufacturer")
        ] == read_dmi_data.call_args_list



# Copyright (C) 2012 Canonical Ltd.
# Copyright (C) 2012, 2013 Hewlett-Packard Development Company, L.P.
# Copyright (C) 2012 Yahoo! Inc.
#
# Author: Scott Moser <scott.moser@canonical.com>
# Author: Juerg Haefliger <juerg.haefliger@hp.com>
# Author: Joshua Harlow <harlowja@yahoo-inc.com>
#
# This file is part of cloud-init. See LICENSE file for license information.

import abc
import logging
import os

from cloudinit import importer, type_utils, util
from cloudinit.settings import FREQUENCIES, PER_ALWAYS, PER_INSTANCE

LOG = logging.getLogger(__name__)

# Used as the content type when a message is not multipart
# and it doesn't contain its own content-type
NOT_MULTIPART_TYPE = "text/x-not-multipart"

# When none is assigned this gets used
OCTET_TYPE = "application/octet-stream"

# Special content types that signal the start and end of processing
CONTENT_END = "__end__"
CONTENT_START = "__begin__"
CONTENT_SIGNALS = [CONTENT_START, CONTENT_END]

# Used when a part-handler type is encountered
# to allow for registration of new types.
PART_CONTENT_TYPES = ["text/part-handler"]
PART_HANDLER_FN_TMPL = "part-handler-%03d"

# For parts without filenames
PART_FN_TPL = "part-%03d"

# Different file beginnings to their content type
INCLUSION_TYPES_MAP = {
    "#include": "text/x-include-url",
    "#include-once": "text/x-include-once-url",
    "#!": "text/x-shellscript",
    "#cloud-config": "text/cloud-config",
    "#part-handler": "text/part-handler",
    "#cloud-boothook": "text/cloud-boothook",
    "#cloud-config-archive": "text/cloud-config-archive",
    "#cloud-config-jsonp": "text/cloud-config-jsonp",
    "## template: jinja": "text/jinja2",
    # Note: for the next 3 entries, the prefix doesn't matter because these
    # are for types that can only be used as part of a MIME message. However,
    # including these entries suppresses warnings during `cloudinit devel
    # make-mime`, which otherwise would require `--force`.
    "text/x-shellscript-per-boot": "text/x-shellscript-per-boot",
    "text/x-shellscript-per-instance": "text/x-shellscript-per-instance",
    "text/x-shellscript-per-once": "text/x-shellscript-per-once",
}

# Sorted longest first
INCLUSION_SRCH = sorted(
    list(INCLUSION_TYPES_MAP.keys()), key=(lambda e: 0 - len(e))
)


class Handler(metaclass=abc.ABCMeta):
    def __init__(self, frequency, version=2):
        self.handler_version = version
        self.frequency = frequency

    def __repr__(self):
        return "%s: [%s]" % (type_utils.obj_name(self), self.list_types())

    def list_types(self):
        # Each subclass must define the supported content prefixes it handles.
        if not hasattr(self, "prefixes"):
            raise NotImplementedError("Missing prefixes subclass attribute")
        else:
            return [
                INCLUSION_TYPES_MAP[prefix]
                for prefix in getattr(self, "prefixes")
            ]

    @abc.abstractmethod
    def handle_part(self, *args, **kwargs):
        raise NotImplementedError()


def run_part(mod, data, filename, payload, frequency, headers):
    mod_freq = mod.frequency
    if not (
        mod_freq == PER_ALWAYS
        or (frequency == PER_INSTANCE and mod_freq == PER_INSTANCE)
    ):
        return
    # Sanity checks on version (should be an int convertible)
    try:
        mod_ver = mod.handler_version
        mod_ver = int(mod_ver)
    except (TypeError, ValueError, AttributeError):
        mod_ver = 1
    content_type = headers["Content-Type"]
    try:
        LOG.debug(
            "Calling handler %s (%s, %s, %s) with frequency %s",
            mod,
            content_type,
            filename,
            mod_ver,
            frequency,
        )
        if mod_ver == 3:
            # Treat as v. 3 which does get a frequency + headers
            mod.handle_part(
                data, content_type, filename, payload, frequency, headers
            )
        elif mod_ver == 2:
            # Treat as v. 2 which does get a frequency
            mod.handle_part(data, content_type, filename, payload, frequency)
        elif mod_ver == 1:
            # Treat as v. 1 which gets no frequency
            mod.handle_part(data, content_type, filename, payload)
        else:
            raise ValueError("Unknown module version %s" % (mod_ver))
    except Exception:
        util.logexc(
            LOG,
            "Failed calling handler %s (%s, %s, %s) with frequency %s",
            mod,
            content_type,
            filename,
            mod_ver,
            frequency,
        )


def call_begin(mod, data, frequency):
    # Create a fake header set
    headers = {
        "Content-Type": CONTENT_START,
    }
    run_part(mod, data, None, None, frequency, headers)


def call_end(mod, data, frequency):
    # Create a fake header set
    headers = {
        "Content-Type": CONTENT_END,
    }
    run_part(mod, data, None, None, frequency, headers)


def walker_handle_handler(pdata, _ctype, _filename, payload):
    curcount = pdata["handlercount"]
    modname = PART_HANDLER_FN_TMPL % (curcount)
    frequency = pdata["frequency"]
    modfname = os.path.join(pdata["handlerdir"], "%s" % (modname))
    if not modfname.endswith(".py"):
        modfname = "%s.py" % (modfname)
    # TODO(harlowja): Check if path exists??
    util.write_file(modfname, payload, 0o600)
    handlers = pdata["handlers"]
    try:
        mod = fixup_handler(importer.import_module(modname))
        call_begin(mod, pdata["data"], frequency)
        # Only register and increment after the above have worked, so we don't
        # register if it fails starting.
        handlers.register(mod, initialized=True)
        pdata["handlercount"] = curcount + 1
    except Exception:
        util.logexc(
            LOG,
            "Failed at registering python file: %s (part handler %s)",
            modfname,
            curcount,
        )


def _extract_first_or_bytes(blob, size):
    # Extract the first line or upto X symbols for text objects
    # Extract first X bytes for binary objects
    try:
        if isinstance(blob, str):
            start = blob.split("\n", 1)[0]
        else:
            # We want to avoid decoding the whole blob (it might be huge)
            # By taking 4*size bytes we guarantee to decode size utf8 chars
            start = blob[: 4 * size].decode(errors="ignore").split("\n", 1)[0]
        if len(start) >= size:
            start = start[:size]
    except UnicodeDecodeError:
        # Bytes array doesn't contain text so return chunk of raw bytes
        start = blob[0:size]
    return start


def _escape_string(text):
    try:
        return text.encode("string_escape")
    except (LookupError, TypeError):
        try:
            # Unicode (and Python 3's str) doesn't support string_escape...
            return text.encode("unicode_escape")
        except TypeError:
            # Give up...
            pass
    except AttributeError:
        # We're in Python3 and received blob as text
        # No escaping is needed because bytes are printed
        # as 'b\xAA\xBB' automatically in Python3
        pass
    return text


def walker_callback(data, filename, payload, headers):
    content_type = headers["Content-Type"]
    if content_type in data.get("excluded"):
        LOG.debug('content_type "%s" is excluded', content_type)
        return

    if content_type in PART_CONTENT_TYPES:
        walker_handle_handler(data, content_type, filename, payload)
        return
    handlers = data["handlers"]
    if content_type in handlers:
        run_part(
            handlers[content_type],
            data["data"],
            filename,
            payload,
            data["frequency"],
            headers,
        )
    elif payload:
        # Extract the first line or 24 bytes for displaying in the log
        start = _extract_first_or_bytes(payload, 24)
        details = "'%s...'" % (_escape_string(start))
        if content_type == NOT_MULTIPART_TYPE:
            LOG.warning(
                "Unhandled non-multipart (%s) userdata: %s",
                content_type,
                details,
            )
        else:
            LOG.warning(
                "Unhandled unknown content-type (%s) userdata: %s",
                content_type,
                details,
            )
    else:
        LOG.debug("Empty payload of type %s", content_type)


# Callback is a function that will be called with
# (data, content_type, filename, payload)
def walk(msg, callback, data):
    partnum = 0
    for part in msg.walk():
        # multipart/* are just containers
        if part.get_content_maintype() == "multipart":
            continue

        ctype = part.get_content_type()
        if ctype is None:
            ctype = OCTET_TYPE

        filename = part.get_filename()
        if not filename:
            filename = PART_FN_TPL % (partnum)

        headers = dict(part)
        LOG.debug(headers)
        headers["Content-Type"] = ctype
        payload = util.fully_decoded_payload(part)
        callback(data, filename, payload, headers)
        partnum = partnum + 1


def fixup_handler(mod, def_freq=PER_INSTANCE):
    if not hasattr(mod, "handler_version"):
        setattr(mod, "handler_version", 1)
    if not hasattr(mod, "frequency"):
        setattr(mod, "frequency", def_freq)
    else:
        freq = mod.frequency
        if freq and freq not in FREQUENCIES:
            LOG.warning("Handler %s has an unknown frequency %s", mod, freq)
    return mod


def type_from_starts_with(payload, default=None):
    try:
        payload_lc = util.decode_binary(payload).lower()
    except UnicodeDecodeError:
        return default
    payload_lc = payload_lc.lstrip()
    for text in INCLUSION_SRCH:
        if payload_lc.startswith(text):
            return INCLUSION_TYPES_MAP[text]
    return default



# This file is part of cloud-init. See LICENSE file for license information.

import calendar
import sys
from datetime import datetime

from cloudinit import atomic_helper, subp, util

stage_to_description = {
    "finished": "finished running cloud-init",
    "init-local": "starting search for local datasources",
    "init-network": "searching for network datasources",
    "init": "searching for network datasources",
    "modules-config": "running config modules",
    "modules-final": "finalizing modules",
    "modules": "running modules for",
    "single": "running single module ",
}

# logger's asctime format
CLOUD_INIT_ASCTIME_FMT = "%Y-%m-%d %H:%M:%S,%f"

# journctl -o short-precise
CLOUD_INIT_JOURNALCTL_FMT = "%b %d %H:%M:%S.%f %Y"

# other
DEFAULT_FMT = "%b %d %H:%M:%S %Y"


def parse_timestamp(timestampstr):
    # default syslog time does not include the current year
    months = [calendar.month_abbr[m] for m in range(1, 13)]
    if timestampstr.split()[0] in months:
        # Aug 29 22:55:26
        FMT = DEFAULT_FMT
        if "." in timestampstr:
            FMT = CLOUD_INIT_JOURNALCTL_FMT
        dt = datetime.strptime(
            timestampstr + " " + str(datetime.now().year), FMT
        )
        timestamp = dt.strftime("%s.%f")
    elif "," in timestampstr:
        # 2016-09-12 14:39:20,839
        dt = datetime.strptime(timestampstr, CLOUD_INIT_ASCTIME_FMT)
        timestamp = dt.strftime("%s.%f")
    else:
        # allow GNU date(1) to handle other formats we don't expect
        # This may throw a ValueError if no GNU date can be found
        timestamp = parse_timestamp_from_date(timestampstr)

    return float(timestamp)


def parse_timestamp_from_date(timestampstr):
    cmd = "date"
    if not util.is_Linux():
        if subp.which("gdate"):
            cmd = "gdate"
        else:
            raise ValueError(
                f"Unable to parse timestamp without GNU date: [{timestampstr}]"
            )
    out, _ = subp.subp([cmd, "+%s.%3N", "-d", timestampstr])
    timestamp = out.strip()
    return float(timestamp)


def parse_ci_logline(line):
    # Stage Starts:
    # Cloud-init v. 0.7.7 running 'init-local' at \
    #               Fri, 02 Sep 2016 19:28:07 +0000. Up 1.0 seconds.
    # Cloud-init v. 0.7.7 running 'init' at \
    #               Fri, 02 Sep 2016 19:28:08 +0000. Up 2.0 seconds.
    # Cloud-init v. 0.7.7 finished at
    # Aug 29 22:55:26 test1 [CLOUDINIT] handlers.py[DEBUG]: \
    #               finish: modules-final: SUCCESS: running modules for final
    # 2016-08-30T21:53:25.972325+00:00 y1 [CLOUDINIT] handlers.py[DEBUG]: \
    #               finish: modules-final: SUCCESS: running modules for final
    #
    # Nov 03 06:51:06.074410 x2 cloud-init[106]: [CLOUDINIT] util.py[DEBUG]: \
    #               Cloud-init v. 0.7.8 running 'init-local' at \
    #               Thu, 03 Nov 2016 06:51:06 +0000. Up 1.0 seconds.
    #
    # 2017-05-22 18:02:01,088 - util.py[DEBUG]: Cloud-init v. 0.7.9 running \
    #         'init-local' at Mon, 22 May 2017 18:02:01 +0000. Up 2.0 seconds.
    #
    # Apr 30 19:39:11 cloud-init[2673]: handlers.py[DEBUG]: start: \
    #          init-local/check-cache: attempting to read from cache [check]

    amazon_linux_2_sep = " cloud-init["
    separators = [" - ", " [CLOUDINIT] ", amazon_linux_2_sep]
    found = False
    for sep in separators:
        if sep in line:
            found = True
            break

    if not found:
        return None

    (timehost, eventstr) = line.split(sep)

    # journalctl -o short-precise
    if timehost.endswith(":"):
        timehost = " ".join(timehost.split()[0:-1])

    if "," in timehost:
        timestampstr, extra = timehost.split(",")
        timestampstr += ",%s" % extra.split()[0]
        if " " in extra:
            hostname = extra.split()[-1]
    else:
        hostname = timehost.split()[-1]
        if sep == amazon_linux_2_sep:
            # This is an Amazon Linux style line, with no hostname and a PID.
            # Use the whole of timehost as timestampstr, and strip off the PID
            # from the start of eventstr.
            timestampstr = timehost.strip()
            eventstr = eventstr.split(maxsplit=1)[1]
        else:
            timestampstr = timehost.split(hostname)[0].strip()
    if "Cloud-init v." in eventstr:
        event_type = "start"
        if "running" in eventstr:
            stage_and_timestamp = eventstr.split("running")[1].lstrip()
            event_name, _ = stage_and_timestamp.split(" at ")
            event_name = event_name.replace("'", "").replace(":", "-")
            if event_name == "init":
                event_name = "init-network"
        else:
            # don't generate a start for the 'finished at' banner
            return None
        event_description = stage_to_description[event_name]
    else:
        (_pymodloglvl, event_type, event_name) = eventstr.split()[0:3]
        event_description = eventstr.split(event_name)[1].strip()

    event = {
        "name": event_name.rstrip(":"),
        "description": event_description,
        "timestamp": parse_timestamp(timestampstr),
        "origin": "cloudinit",
        "event_type": event_type.rstrip(":"),
    }
    if event["event_type"] == "finish":
        result = event_description.split(":")[0]
        desc = event_description.split(result)[1].lstrip(":").strip()
        event["result"] = result
        event["description"] = desc.strip()

    return event


def dump_events(cisource=None, rawdata=None):
    events = []
    event = None
    CI_EVENT_MATCHES = ["start:", "finish:", "Cloud-init v."]

    if not any([cisource, rawdata]):
        raise ValueError("Either cisource or rawdata parameters are required")

    if rawdata:
        data = rawdata.splitlines()
    else:
        data = cisource.readlines()

    for line in data:
        for match in CI_EVENT_MATCHES:
            if match in line:
                try:
                    event = parse_ci_logline(line)
                except ValueError:
                    sys.stderr.write("Skipping invalid entry\n")
                if event:
                    events.append(event)

    return events, data


def main():
    if len(sys.argv) > 1:
        cisource = open(sys.argv[1])
    else:
        cisource = sys.stdin

    return atomic_helper.json_dumps(dump_events(cisource))


if __name__ == "__main__":
    print(main())



# This file is part of cloud-init. See LICENSE file for license information.
"""Tests cc_fan module"""

import pytest

from cloudinit.config.schema import (
    SchemaValidationError,
    get_schema,
    validate_cloudconfig_schema,
)
from tests.unittests.helpers import skipUnlessJsonSchema


class TestFanSchema:
    @pytest.mark.parametrize(
        "config, error_msg",
        (
            ({"fan": {"config": ["a", "b"]}}, "is not of type 'string'"),
            (
                {"fan": {"config_path": "/a/b"}},
                "'config' is a required property",
            ),
        ),
    )
    @skipUnlessJsonSchema()
    def test_schema_validation(self, config, error_msg):
        """Assert expected schema validation and error messages."""
        schema = get_schema()
        if error_msg is None:
            validate_cloudconfig_schema(config, schema, strict=True)
        else:
            with pytest.raises(SchemaValidationError, match=error_msg):
                validate_cloudconfig_schema(config, schema, strict=True)



# This file is part of cloud-init. See LICENSE file for license information.
import datetime
import functools
import logging
import os
import shutil
import subprocess
import sys
from contextlib import contextmanager
from pathlib import Path
from tarfile import TarFile
from typing import Dict, Generator, Iterator, List, Type

import pytest
from pycloudlib.lxd.instance import LXDInstance

from tests.integration_tests import integration_settings
from tests.integration_tests.clouds import (
    AzureCloud,
    Ec2Cloud,
    GceCloud,
    IbmCloud,
    IntegrationCloud,
    LxdContainerCloud,
    LxdVmCloud,
    OciCloud,
    OpenstackCloud,
    QemuCloud,
    _LxdIntegrationCloud,
)
from tests.integration_tests.instances import (
    CloudInitSource,
    IntegrationInstance,
)

log = logging.getLogger("integration_testing")
log.addHandler(logging.StreamHandler(sys.stdout))
log.setLevel(logging.INFO)

platforms: Dict[str, Type[IntegrationCloud]] = {
    "ec2": Ec2Cloud,
    "gce": GceCloud,
    "azure": AzureCloud,
    "oci": OciCloud,
    "ibm": IbmCloud,
    "lxd_container": LxdContainerCloud,
    "lxd_vm": LxdVmCloud,
    "qemu": QemuCloud,
    "openstack": OpenstackCloud,
}
os_list = ["ubuntu"]

session_start_time = datetime.datetime.now().strftime("%y%m%d%H%M%S")


def pytest_runtest_setup(item):
    """Skip tests on unsupported clouds.

    A test can take any number of marks to specify the platforms it can
    run on. If a platform(s) is specified and we're not running on that
    platform, then skip the test. If platform specific marks are not
    specified, then we assume the test can be run anywhere.
    """
    test_marks = [mark.name for mark in item.iter_markers()]
    if "unstable" in test_marks and not integration_settings.RUN_UNSTABLE:
        pytest.skip("Test marked unstable. Manually remove mark to run it")


# disable_subp_usage is defined at a higher level, but we don't
# want it applied here
@pytest.fixture()
def disable_subp_usage(request):
    pass


@pytest.fixture(scope="session")
def session_cloud() -> Generator[IntegrationCloud, None, None]:
    if integration_settings.PLATFORM not in platforms.keys():
        raise ValueError(
            f"{integration_settings.PLATFORM} is an invalid PLATFORM "
            f"specified in settings. Must be one of {list(platforms.keys())}"
        )

    cloud = platforms[integration_settings.PLATFORM]()
    cloud.emit_settings_to_log()
    yield cloud
    cloud.destroy()


def get_validated_source(
    session_cloud: IntegrationCloud,
    source=integration_settings.CLOUD_INIT_SOURCE,
) -> CloudInitSource:
    if source == "NONE":
        return CloudInitSource.NONE
    elif source == "IN_PLACE":
        if session_cloud.datasource not in ["lxd_container", "lxd_vm"]:
            raise ValueError(
                "IN_PLACE as CLOUD_INIT_SOURCE only works for LXD"
            )
        return CloudInitSource.IN_PLACE
    elif source == "PROPOSED":
        return CloudInitSource.PROPOSED
    elif source.startswith("ppa:"):
        return CloudInitSource.PPA
    elif os.path.isfile(str(source)):
        return CloudInitSource.DEB_PACKAGE
    elif source == "UPGRADE":
        return CloudInitSource.UPGRADE
    raise ValueError(f"Invalid value for CLOUD_INIT_SOURCE setting: {source}")


@pytest.fixture(scope="session")
def setup_image(session_cloud: IntegrationCloud, request):
    """Setup the target environment with the correct version of cloud-init.

    So we can launch instances / run tests with the correct image
    """
    source = get_validated_source(session_cloud)
    if not (
        source.installs_new_version() or integration_settings.INCLUDE_COVERAGE
    ):
        return
    log.info("Setting up source image")
    client = session_cloud.launch()
    if source.installs_new_version():
        log.info("Installing cloud-init from %s", source.name)
        client.install_new_cloud_init(source)
    if integration_settings.INCLUDE_COVERAGE:
        log.info("Installing coverage")
        client.install_coverage()
    # All done customizing the image, so snapshot it and make it global
    snapshot_id = client.snapshot()
    client.cloud.snapshot_id = snapshot_id
    # Even if we're keeping instances, we don't want to keep this
    # one around as it was just for image creation
    client.destroy()
    log.info("Done with environment setup")

    # For some reason a yield here raises a
    # ValueError: setup_image did not yield a value
    # during setup so use a finalizer instead.
    request.addfinalizer(session_cloud.delete_snapshot)


def _collect_logs(instance: IntegrationInstance, log_dir: Path):
    instance.execute(
        "cloud-init collect-logs -u -t /var/tmp/cloud-init.tar.gz"
    )
    log.info("Writing logs to %s", log_dir)

    tarball_path = log_dir / "cloud-init.tar.gz"
    try:
        instance.pull_file("/var/tmp/cloud-init.tar.gz", tarball_path)
    except Exception as e:
        log.error("Failed to pull logs: %s", e)
        return

    tarball = TarFile.open(str(tarball_path))
    tarball.extractall(path=str(log_dir))
    tarball_path.unlink()


def _collect_coverage(instance: IntegrationInstance, log_dir: Path):
    log.info("Writing coverage report to %s", log_dir)
    try:
        instance.pull_file("/.coverage", log_dir / ".coverage")
    except Exception as e:
        log.error("Failed to pull coverage for: %s", e)


def _setup_artifact_paths(node_id: str):
    parent_dir = Path(integration_settings.LOCAL_LOG_PATH, session_start_time)

    node_id_path = Path(
        node_id.replace(
            ".py", ""
        )  # Having a directory with '.py' would be weird
        .replace("::", os.path.sep)  # Turn classes/tests into paths
        .replace("[", "-")  # For parametrized names
        .replace("]", "")  # For parameterized names
    )
    log_dir = parent_dir / node_id_path

    # Create log dir if not exists
    if not log_dir.exists():
        log_dir.mkdir(parents=True)

    # Add a symlink to the latest log output directory
    last_symlink = Path(integration_settings.LOCAL_LOG_PATH) / "last"
    if os.path.islink(last_symlink):
        os.unlink(last_symlink)
    os.symlink(parent_dir, last_symlink)
    return log_dir


def _collect_artifacts(
    instance: IntegrationInstance, node_id: str, test_failed: bool
):
    """Collect artifacts from remote instance.

    Args:
        instance: The current IntegrationInstance to collect artifacts from
        node_id: The pytest representation of this test, E.g.:
            tests/integration_tests/test_example.py::TestExample.test_example
        test_failed: If test failed or not
    """
    should_collect_logs = integration_settings.COLLECT_LOGS == "ALWAYS" or (
        integration_settings.COLLECT_LOGS == "ON_ERROR" and test_failed
    )
    should_collect_coverage = integration_settings.INCLUDE_COVERAGE
    if not (should_collect_logs or should_collect_coverage):
        return

    log_dir = _setup_artifact_paths(node_id)

    if should_collect_logs:
        _collect_logs(instance, log_dir)

    if should_collect_coverage:
        _collect_coverage(instance, log_dir)


@contextmanager
def _client(
    request, fixture_utils, session_cloud: IntegrationCloud
) -> Iterator[IntegrationInstance]:
    """Fixture implementation for the client fixtures.

    Launch the dynamic IntegrationClient instance using any provided
    userdata, yield to the test, then cleanup
    """
    getter = functools.partial(
        fixture_utils.closest_marker_first_arg_or, request, default=None
    )
    user_data = getter("user_data")
    name = getter("instance_name")
    lxd_config_dict = getter("lxd_config_dict")
    lxd_setup = getter("lxd_setup")
    lxd_use_exec = fixture_utils.closest_marker_args_or(
        request, "lxd_use_exec", None
    )
    launch_kwargs = {}
    if name is not None:
        launch_kwargs["name"] = name
    if lxd_config_dict is not None:
        if not isinstance(session_cloud, _LxdIntegrationCloud):
            pytest.skip("lxd_config_dict requires LXD")
        launch_kwargs["config_dict"] = lxd_config_dict
    if lxd_use_exec is not None:
        if not isinstance(session_cloud, _LxdIntegrationCloud):
            pytest.skip("lxd_use_exec requires LXD")
        launch_kwargs["execute_via_ssh"] = False
    local_launch_kwargs = {}
    if lxd_setup is not None:
        if not isinstance(session_cloud, _LxdIntegrationCloud):
            pytest.skip("lxd_setup requires LXD")
        local_launch_kwargs["lxd_setup"] = lxd_setup

    with session_cloud.launch(
        user_data=user_data, launch_kwargs=launch_kwargs, **local_launch_kwargs
    ) as instance:
        if lxd_use_exec is not None and isinstance(
            instance.instance, LXDInstance
        ):
            # Existing instances are not affected by the launch kwargs, so
            # ensure it here; we still need the launch kwarg so waiting works
            instance.instance.execute_via_ssh = False
        previous_failures = request.session.testsfailed
        yield instance
        test_failed = request.session.testsfailed - previous_failures > 0
        _collect_artifacts(instance, request.node.nodeid, test_failed)


@pytest.fixture
def client(
    request, fixture_utils, session_cloud, setup_image
) -> Iterator[IntegrationInstance]:
    """Provide a client that runs for every test."""
    with _client(request, fixture_utils, session_cloud) as client:
        yield client


@pytest.fixture(scope="module")
def module_client(
    request, fixture_utils, session_cloud, setup_image
) -> Iterator[IntegrationInstance]:
    """Provide a client that runs once per module."""
    with _client(request, fixture_utils, session_cloud) as client:
        yield client


@pytest.fixture(scope="class")
def class_client(
    request, fixture_utils, session_cloud, setup_image
) -> Iterator[IntegrationInstance]:
    """Provide a client that runs once per class."""
    with _client(request, fixture_utils, session_cloud) as client:
        yield client


def pytest_assertrepr_compare(op, left, right):
    """Custom integration test assertion explanations.

    See
    https://docs.pytest.org/en/stable/assert.html#defining-your-own-explanation-for-failed-assertions
    for pytest's documentation.
    """
    if op == "not in" and isinstance(left, str) and isinstance(right, str):
        # This stanza emits an improved assertion message if we're testing for
        # the presence of a string within a cloud-init log: it will report only
        # the specific lines containing the string (instead of the full log,
        # the default behaviour).
        potential_log_lines = right.splitlines()
        first_line = potential_log_lines[0]
        if "DEBUG" in first_line and "Cloud-init" in first_line:
            # We are looking at a cloud-init log, so just pick out the relevant
            # lines
            found_lines = [
                line for line in potential_log_lines if left in line
            ]
            return [
                '"{}" not in cloud-init.log string; unexpectedly found on'
                " these lines:".format(left)
            ] + found_lines


def pytest_configure(config):
    """Perform initial configuration, before the test runs start.

    This hook is only called if integration tests are being executed, so we can
    use it to configure defaults for integration testing that differ from the
    rest of the tests in the codebase.

    See
    https://docs.pytest.org/en/latest/reference.html#_pytest.hookspec.pytest_configure
    for pytest's documentation.
    """
    if "log_cli_level" in config.option and not config.option.log_cli_level:
        # If log_cli_level is available in this version of pytest and not set
        # to anything, set it to INFO.
        config.option.log_cli_level = "INFO"


def _copy_coverage_files(parent_dir: Path) -> List[Path]:
    combined_files = []
    for dirpath in parent_dir.rglob("*"):
        if (dirpath / ".coverage").exists():
            # Construct the new filename
            relative_dir = dirpath.relative_to(parent_dir)
            new_filename = ".coverage." + str(relative_dir).replace(
                os.sep, "-"
            )
            new_filepath = parent_dir / new_filename

            # Copy the file
            shutil.copy(dirpath / ".coverage", new_filepath)
            combined_files.append(new_filepath)
    return combined_files


def _generate_coverage_report() -> None:
    log.info("Generating coverage report")
    parent_dir = Path(integration_settings.LOCAL_LOG_PATH, session_start_time)
    coverage_files = _copy_coverage_files(parent_dir)
    subprocess.run(
        ["coverage", "combine"] + [str(f) for f in coverage_files],
        check=True,
        cwd=str(parent_dir),
        stdout=subprocess.DEVNULL,
    )
    html_dir = parent_dir / "html"
    html_dir.mkdir()
    subprocess.run(
        [
            "coverage",
            "html",
            f"--data-file={parent_dir / '.coverage'}",
            f"--directory={html_dir}",
            "--ignore-errors",
        ],
        check=True,
        stdout=subprocess.DEVNULL,
    )
    log.info("Coverage report generated")


def pytest_sessionfinish(session, exitstatus) -> None:
    if not integration_settings.INCLUDE_COVERAGE:
        return
    _generate_coverage_report()



import os
import sys
from importlib import reload

import pytest

from cloudinit import apport
from tests.unittests.helpers import mock

M_PATH = "cloudinit.apport."


class TestApport:
    @pytest.mark.parametrize(
        "instance_data,choice_idx,expected_report",
        (
            pytest.param(
                '{"v1": {"cloud_name": "mycloud"}}',
                None,
                {},
                id="v1_cloud_name_exists",
            ),
            pytest.param(
                '{"v1": {"cloud_id": "invalid"}}',
                1,
                {"CloudName": "Azure"},
                id="v1_no_cloud_name_present",
            ),
            pytest.param("{}", 0, {"CloudName": "AliYun"}, id="no_v1_key"),
            pytest.param(
                "{", 22, {"CloudName": "Oracle"}, id="not_valid_json"
            ),
        ),
    )
    def test_attach_cloud_info(
        self, instance_data, choice_idx, expected_report, mocker, paths
    ):
        """Prompt for cloud name when instance-data.json is not-json/absent."""
        mocker.patch(M_PATH + "read_cfg_paths", return_value=paths)
        instance_data_file = paths.get_runpath("instance_data")
        if instance_data is None:
            assert not os.path.exists(instance_data_file)
        else:
            with open(instance_data_file, "w") as stream:
                stream.write(instance_data)
        ui = mock.Mock()
        ui.yesno.return_value = True
        ui.choice.return_value = (choice_idx, "")
        report = {}
        apport.attach_cloud_info(report, ui)
        if choice_idx is not None:
            assert ui.choice.call_count == 1
            assert report["CloudName"] == apport.KNOWN_CLOUD_NAMES[choice_idx]
        else:
            assert ui.choice.call_count == 0

    def test_attach_user_data(self, mocker, paths):
        user_data_file = paths.get_ipath_cur("userdata_raw")
        ui = mock.Mock()
        ui.yesno.return_value = True
        report = object()
        m_hookutils = mock.Mock()

        with mock.patch.dict(sys.modules, {"apport.hookutils": m_hookutils}):
            reload(sys.modules["cloudinit.apport"])
            mocker.patch(M_PATH + "read_cfg_paths", return_value=paths)
        apport.attach_user_data(report, ui)
        assert [
            mock.call(report, user_data_file, "user_data.txt"),
        ] == apport.attach_file.call_args_list
        assert [
            mock.call(
                report,
                "/var/log/installer/autoinstall-user-data",
                "AutoInstallUserData",
            ),
            mock.call(report, "/autoinstall.yaml", "AutoInstallYAML"),
            mock.call(
                report,
                "/etc/cloud/cloud.cfg.d/99-installer.cfg",
                "InstallerCloudCfg",
            ),
        ] == apport.attach_file_if_exists.call_args_list

    @pytest.mark.parametrize(
        "report,tags",
        (
            ({"Irrelevant": "."}, ""),
            ({"UdiLog": "."}, "ubuntu-desktop-installer"),
            ({"CurtinError": ".", "SubiquityLog": "."}, "curtin subiquity"),
            (
                {
                    "UdiLog": ".",
                    "JournalErrors": "...Breaking ordering cycle...",
                },
                "systemd-ordering ubuntu-desktop-installer",
            ),
        ),
    )
    def test_add_bug_tags_assigns_proper_tags(self, report, tags):
        """Tags are assigned based on non-empty project report key values."""

        apport.add_bug_tags(report)
        assert report.get("Tags", "") == tags

    @mock.patch(M_PATH + "os.path.exists", return_value=True)
    def test_attach_ubuntu_pro_info(self, m_exists):
        m_hookutils = mock.Mock()
        with mock.patch.dict(sys.modules, {"apport.hookutils": m_hookutils}):
            reload(sys.modules["cloudinit.apport"])
            report = {}
            apport.attach_ubuntu_pro_info(report)

        assert [
            mock.call(report, "/var/log/ubuntu-advantage.log"),
        ] == m_hookutils.attach_file_if_exists.call_args_list
        assert report.get("Tags", "") == "ubuntu-pro"

    @mock.patch(M_PATH + "os.path.exists", return_value=False)
    def test_attach_ubuntu_pro_info_log_non_present(self, m_exists):
        m_hookutils = mock.Mock()
        with mock.patch.dict(sys.modules, {"apport.hookutils": m_hookutils}):
            reload(sys.modules["cloudinit.apport"])
            report = {}
            apport.attach_ubuntu_pro_info(report)

        assert [
            mock.call(report, "/var/log/ubuntu-advantage.log"),
        ] == m_hookutils.attach_file_if_exists.call_args_list
        assert report.get("Tags", "") == ""



# Copyright (C) 2016 Canonical Ltd.
#
# Author: Wesley Wiedenmeier <wesley.wiedenmeier@canonical.com>
#
# This file is part of cloud-init. See LICENSE file for license information.

"""LXD: configure lxd with ``lxd init`` and optionally lxd-bridge"""

import logging
import os
from textwrap import dedent
from typing import List, Tuple

from cloudinit import safeyaml, subp, util
from cloudinit.cloud import Cloud
from cloudinit.config import Config
from cloudinit.config.schema import MetaSchema, get_meta_doc
from cloudinit.settings import PER_INSTANCE

LOG = logging.getLogger(__name__)

_DEFAULT_NETWORK_NAME = "lxdbr0"


MODULE_DESCRIPTION = """\
This module configures lxd with user specified options using ``lxd init``.
If lxd is not present on the system but lxd configuration is provided, then
lxd will be installed. If the selected storage backend userspace utility is
not installed, it will be installed. If network bridge configuration is
provided, then lxd-bridge will be configured accordingly.
"""

distros = ["ubuntu"]

meta: MetaSchema = {
    "id": "cc_lxd",
    "name": "LXD",
    "title": "Configure LXD with ``lxd init`` and optionally lxd-bridge",
    "description": MODULE_DESCRIPTION,
    "distros": distros,
    "examples": [
        dedent(
            """\
            # Simplest working directory backed LXD configuration
            lxd:
              init:
                storage_backend: dir
            """
        ),
        dedent(
            """\
            # LXD init showcasing cloud-init's LXD config options
            lxd:
              init:
                network_address: 0.0.0.0
                network_port: 8443
                storage_backend: zfs
                storage_pool: datapool
                storage_create_loop: 10
              bridge:
                mode: new
                mtu: 1500
                name: lxdbr0
                ipv4_address: 10.0.8.1
                ipv4_netmask: 24
                ipv4_dhcp_first: 10.0.8.2
                ipv4_dhcp_last: 10.0.8.3
                ipv4_dhcp_leases: 250
                ipv4_nat: true
                ipv6_address: fd98:9e0:3744::1
                ipv6_netmask: 64
                ipv6_nat: true
                domain: lxd
            """
        ),
        dedent(
            """\
            # For more complex non-iteractive LXD configuration of networks,
            # storage_pools, profiles, projects, clusters and core config,
            # `lxd:preseed` config will be passed as stdin to the command:
            #  lxd init --preseed
            # See https://documentation.ubuntu.com/lxd/en/latest/howto/initialize/#non-interactive-configuration or
            # run: lxd init --dump to see viable preseed YAML allowed.
            #
            # Preseed settings configuring the LXD daemon for HTTPS connections
            # on 192.168.1.1 port 9999, a nested profile which allows for
            # LXD nesting on containers and a limited project allowing for
            # RBAC approach when defining behavior for sub projects.
            lxd:
              preseed: |
                config:
                  core.https_address: 192.168.1.1:9999
                networks:
                  - config:
                      ipv4.address: 10.42.42.1/24
                      ipv4.nat: true
                      ipv6.address: fd42:4242:4242:4242::1/64
                      ipv6.nat: true
                    description: ""
                    name: lxdbr0
                    type: bridge
                    project: default
                storage_pools:
                  - config:
                      size: 5GiB
                      source: /var/snap/lxd/common/lxd/disks/default.img
                    description: ""
                    name: default
                    driver: zfs
                profiles:
                  - config: {}
                    description: Default LXD profile
                    devices:
                      eth0:
                        name: eth0
                        network: lxdbr0
                        type: nic
                      root:
                        path: /
                        pool: default
                        type: disk
                    name: default
                  - config: {}
                    security.nesting: true
                    devices:
                      eth0:
                        name: eth0
                        network: lxdbr0
                        type: nic
                      root:
                        path: /
                        pool: default
                        type: disk
                    name: nested
                projects:
                  - config:
                      features.images: true
                      features.networks: true
                      features.profiles: true
                      features.storage.volumes: true
                    description: Default LXD project
                    name: default
                  - config:
                      features.images: false
                      features.networks: true
                      features.profiles: false
                      features.storage.volumes: false
                    description: Limited Access LXD project
                    name: limited


            """  # noqa: E501
        ),
    ],
    "frequency": PER_INSTANCE,
    "activate_by_schema_keys": ["lxd"],
}

__doc__ = get_meta_doc(meta)


def supplemental_schema_validation(
    init_cfg: dict, bridge_cfg: dict, preseed_str: str
):
    """Validate user-provided lxd network and bridge config option values.

    @raises: ValueError describing invalid values provided.
    """
    errors = []
    if not isinstance(init_cfg, dict):
        errors.append(
            f"lxd.init config must be a dictionary. found a"
            f" '{type(init_cfg).__name__}'",
        )

    if not isinstance(bridge_cfg, dict):
        errors.append(
            f"lxd.bridge config must be a dictionary. found a"
            f" '{type(bridge_cfg).__name__}'",
        )

    if not isinstance(preseed_str, str):
        errors.append(
            f"lxd.preseed config must be a string. found a"
            f" '{type(preseed_str).__name__}'",
        )
    if preseed_str and (init_cfg or bridge_cfg):
        incompat_cfg = ["lxd.init"] if init_cfg else []
        incompat_cfg += ["lxd.bridge"] if bridge_cfg else []

        errors.append(
            "Unable to configure LXD. lxd.preseed config can not be provided"
            f" with key(s): {', '.join(incompat_cfg)}"
        )
    if errors:
        raise ValueError(". ".join(errors))


def handle(name: str, cfg: Config, cloud: Cloud, args: list) -> None:
    # Get config
    lxd_cfg = cfg.get("lxd")
    if not lxd_cfg:
        LOG.debug(
            "Skipping module named %s, not present or disabled by cfg", name
        )
        return
    if not isinstance(lxd_cfg, dict):
        raise ValueError(
            f"lxd config must be a dictionary. found a"
            f" '{type(lxd_cfg).__name__}'"
        )

    util.wait_for_snap_seeded(cloud)
    # Grab the configuration
    init_cfg = lxd_cfg.get("init", {})
    preseed_str = lxd_cfg.get("preseed", "")
    bridge_cfg = lxd_cfg.get("bridge", {})
    supplemental_schema_validation(init_cfg, bridge_cfg, preseed_str)

    if not subp.which("lxd"):
        try:
            subp.subp(["snap", "install", "lxd"])
        except subp.ProcessExecutionError as e:
            raise RuntimeError(
                "Failed to install lxd from snap: %s" % e
            ) from e
    packages = get_required_packages(init_cfg, preseed_str)
    if len(packages):
        try:
            cloud.distro.install_packages(packages)
        except subp.ProcessExecutionError as exc:
            LOG.warning("failed to install packages %s: %s", packages, exc)
            return

    subp.subp(["lxd", "waitready", "--timeout=300"])
    if preseed_str:
        subp.subp(["lxd", "init", "--preseed"], data=preseed_str)
        return
    # Set up lxd if init config is given
    if init_cfg:
        # type is known, number of elements is not
        # in the case of the ubuntu+lvm backend workaround
        init_keys: Tuple[str, ...] = (
            "network_address",
            "network_port",
            "storage_backend",
            "storage_create_device",
            "storage_create_loop",
            "storage_pool",
            "trust_password",
        )

        # Bug https://bugs.launchpad.net/ubuntu/+source/linux-kvm/+bug/1982780
        kernel = util.system_info()["uname"][2]
        if init_cfg["storage_backend"] == "lvm" and not os.path.exists(
            f"/lib/modules/{kernel}/kernel/drivers/md/dm-thin-pool.ko"
        ):
            LOG.warning(
                "cloud-init doesn't use thinpool by default on Ubuntu due to "
                "LP #1982780. This behavior will change in the future.",
            )
            subp.subp(
                [
                    "lxc",
                    "storage",
                    "create",
                    "default",
                    "lvm",
                    "lvm.use_thinpool=false",
                ]
            )

            # Since we're manually setting use_thinpool=false
            # filter it from the lxd init commands, don't configure
            # storage twice
            init_keys = tuple(
                key for key in init_keys if key != "storage_backend"
            )

        cmd = ["lxd", "init", "--auto"]
        for k in init_keys:
            if init_cfg.get(k):
                cmd.extend(
                    ["--%s=%s" % (k.replace("_", "-"), str(init_cfg[k]))]
                )
        subp.subp(cmd)

    # Set up lxd-bridge if bridge config is given
    dconf_comm = "debconf-communicate"
    if bridge_cfg:
        net_name = bridge_cfg.get("name", _DEFAULT_NETWORK_NAME)
        if os.path.exists("/etc/default/lxd-bridge") and subp.which(
            dconf_comm
        ):
            # Bridge configured through packaging

            debconf = bridge_to_debconf(bridge_cfg)

            # Update debconf database
            try:
                LOG.debug("Setting lxd debconf via %s", dconf_comm)
                data = (
                    "\n".join(
                        ["set %s %s" % (k, v) for k, v in debconf.items()]
                    )
                    + "\n"
                )
                subp.subp(["debconf-communicate"], data=data)
            except Exception:
                util.logexc(
                    LOG, "Failed to run '%s' for lxd with" % dconf_comm
                )

            # Remove the existing configuration file (forces re-generation)
            util.del_file("/etc/default/lxd-bridge")

            # Run reconfigure
            LOG.debug("Running dpkg-reconfigure for lxd")
            subp.subp(["dpkg-reconfigure", "lxd", "--frontend=noninteractive"])
        else:
            # Built-in LXD bridge support
            cmd_create, cmd_attach = bridge_to_cmd(bridge_cfg)
            maybe_cleanup_default(
                net_name=net_name,
                did_init=bool(init_cfg),
                create=bool(cmd_create),
                attach=bool(cmd_attach),
            )
            if cmd_create:
                LOG.debug("Creating lxd bridge: %s", " ".join(cmd_create))
                _lxc(cmd_create)

            if cmd_attach:
                LOG.debug(
                    "Setting up default lxd bridge: %s", " ".join(cmd_attach)
                )
                _lxc(cmd_attach)

    elif bridge_cfg:
        raise RuntimeError(
            "Unable to configure lxd bridge without %s." + dconf_comm
        )


def bridge_to_debconf(bridge_cfg):
    debconf = {}

    if bridge_cfg.get("mode") == "none":
        debconf["lxd/setup-bridge"] = "false"
        debconf["lxd/bridge-name"] = ""

    elif bridge_cfg.get("mode") == "existing":
        debconf["lxd/setup-bridge"] = "false"
        debconf["lxd/use-existing-bridge"] = "true"
        debconf["lxd/bridge-name"] = bridge_cfg.get("name")

    elif bridge_cfg.get("mode") == "new":
        debconf["lxd/setup-bridge"] = "true"
        if bridge_cfg.get("name"):
            debconf["lxd/bridge-name"] = bridge_cfg.get("name")

        if bridge_cfg.get("ipv4_address"):
            debconf["lxd/bridge-ipv4"] = "true"
            debconf["lxd/bridge-ipv4-address"] = bridge_cfg.get("ipv4_address")
            debconf["lxd/bridge-ipv4-netmask"] = bridge_cfg.get("ipv4_netmask")
            debconf["lxd/bridge-ipv4-dhcp-first"] = bridge_cfg.get(
                "ipv4_dhcp_first"
            )
            debconf["lxd/bridge-ipv4-dhcp-last"] = bridge_cfg.get(
                "ipv4_dhcp_last"
            )
            debconf["lxd/bridge-ipv4-dhcp-leases"] = bridge_cfg.get(
                "ipv4_dhcp_leases"
            )
            debconf["lxd/bridge-ipv4-nat"] = bridge_cfg.get("ipv4_nat", "true")

        if bridge_cfg.get("ipv6_address"):
            debconf["lxd/bridge-ipv6"] = "true"
            debconf["lxd/bridge-ipv6-address"] = bridge_cfg.get("ipv6_address")
            debconf["lxd/bridge-ipv6-netmask"] = bridge_cfg.get("ipv6_netmask")
            debconf["lxd/bridge-ipv6-nat"] = bridge_cfg.get(
                "ipv6_nat", "false"
            )

        if bridge_cfg.get("domain"):
            debconf["lxd/bridge-domain"] = bridge_cfg.get("domain")

    else:
        raise RuntimeError('invalid bridge mode "%s"' % bridge_cfg.get("mode"))

    return debconf


def bridge_to_cmd(bridge_cfg):
    if bridge_cfg.get("mode") == "none":
        return None, None

    bridge_name = bridge_cfg.get("name", _DEFAULT_NETWORK_NAME)
    cmd_create = []
    cmd_attach = ["network", "attach-profile", bridge_name, "default", "eth0"]

    if bridge_cfg.get("mode") == "existing":
        return None, cmd_attach

    if bridge_cfg.get("mode") != "new":
        raise RuntimeError('invalid bridge mode "%s"' % bridge_cfg.get("mode"))

    cmd_create = ["network", "create", bridge_name]

    if bridge_cfg.get("ipv4_address") and bridge_cfg.get("ipv4_netmask"):
        cmd_create.append(
            "ipv4.address=%s/%s"
            % (bridge_cfg.get("ipv4_address"), bridge_cfg.get("ipv4_netmask"))
        )

        if bridge_cfg.get("ipv4_nat", "true") == "true":
            cmd_create.append("ipv4.nat=true")

        if bridge_cfg.get("ipv4_dhcp_first") and bridge_cfg.get(
            "ipv4_dhcp_last"
        ):
            dhcp_range = "%s-%s" % (
                bridge_cfg.get("ipv4_dhcp_first"),
                bridge_cfg.get("ipv4_dhcp_last"),
            )
            cmd_create.append("ipv4.dhcp.ranges=%s" % dhcp_range)
    else:
        cmd_create.append("ipv4.address=none")

    if bridge_cfg.get("ipv6_address") and bridge_cfg.get("ipv6_netmask"):
        cmd_create.append(
            "ipv6.address=%s/%s"
            % (bridge_cfg.get("ipv6_address"), bridge_cfg.get("ipv6_netmask"))
        )

        if bridge_cfg.get("ipv6_nat") == "true":
            cmd_create.append("ipv6.nat=true")

    else:
        cmd_create.append("ipv6.address=none")

    if bridge_cfg.get("domain"):
        cmd_create.append("dns.domain=%s" % bridge_cfg.get("domain"))

    # if the default schema value is passed (-1) don't pass arguments
    # to LXD. Use LXD defaults unless user manually sets a number
    mtu = bridge_cfg.get("mtu", -1)
    if mtu != -1:
        cmd_create.append(f"bridge.mtu={mtu}")

    return cmd_create, cmd_attach


def _lxc(cmd):
    env = {
        "LC_ALL": "C",
        "HOME": os.environ.get("HOME", "/root"),
        "USER": os.environ.get("USER", "root"),
    }
    subp.subp(["lxc"] + list(cmd) + ["--force-local"], update_env=env)


def maybe_cleanup_default(
    net_name, did_init, create, attach, profile="default", nic_name="eth0"
):
    """Newer versions of lxc (3.0.1+) create a lxdbr0 network when
    'lxd init --auto' is run.  Older versions did not.

    By removing any that lxd-init created, we simply leave the add/attach
    code intact.

    https://github.com/canonical/lxd/issues/4649"""
    if net_name != _DEFAULT_NETWORK_NAME or not did_init:
        return

    fail_assume_enoent = "failed. Assuming it did not exist."
    succeeded = "succeeded."
    if create:
        msg = "Detach of lxd network '%s' from profile '%s' %s"
        try:
            _lxc(["network", "detach-profile", net_name, profile])
            LOG.debug(msg, net_name, profile, succeeded)
        except subp.ProcessExecutionError as e:
            if e.exit_code != 1:
                raise e
            LOG.debug(msg, net_name, profile, fail_assume_enoent)
        else:
            msg = "Deletion of lxd network '%s' %s"
            _lxc(["network", "delete", net_name])
            LOG.debug(msg, net_name, succeeded)

    if attach:
        msg = "Removal of device '%s' from profile '%s' %s"
        try:
            _lxc(["profile", "device", "remove", profile, nic_name])
            LOG.debug(msg, nic_name, profile, succeeded)
        except subp.ProcessExecutionError as e:
            if e.exit_code != 1:
                raise e
            LOG.debug(msg, nic_name, profile, fail_assume_enoent)


def get_required_packages(init_cfg: dict, preseed_str: str) -> List[str]:
    """identify required packages for install"""
    packages = []
    # binary for pool creation must be available for the requested backend:
    # zfs, lvcreate, mkfs.btrfs
    storage_drivers: List[str] = []
    preseed_cfg: dict = {}
    if "storage_backend" in init_cfg:
        storage_drivers.append(init_cfg["storage_backend"])
    if preseed_str and "storage_pools" in preseed_str:
        # Assume correct YAML preseed format
        try:
            preseed_cfg = safeyaml.load(preseed_str)
        except (safeyaml.YAMLError, TypeError, ValueError):
            LOG.warning(
                "lxd.preseed string value is not YAML. "
                " Unable to determine required storage driver packages to"
                " support storage_pools config."
            )
    for storage_pool in preseed_cfg.get("storage_pools", []):
        if storage_pool.get("driver"):
            storage_drivers.append(storage_pool["driver"])
    if "zfs" in storage_drivers and not subp.which("zfs"):
        packages.append("zfsutils-linux")
    if "lvm" in storage_drivers and not subp.which("lvcreate"):
        packages.append("lvm2")
    if "btrfs" in storage_drivers and not subp.which("mkfs.btrfs"):
        packages.append("btrfs-progs")
    return packages
